#
# Copyright 2021-2023 Canonical Ltd.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#

"""Executor module."""

from __future__ import annotations

import contextlib
import hashlib
import logging
import re
from abc import ABC, abstractmethod
from os import PathLike
from typing import (
    IO,
    TYPE_CHECKING,
    Any,
    Literal,
    TypeAlias,
    overload,
)

from typing_extensions import Buffer

import craft_providers.util.temp_paths

if TYPE_CHECKING:
    import io
    import pathlib
    import subprocess
    from collections.abc import Callable, Collection, Generator, Iterable

    from craft_providers.errors import ProviderError

logger = logging.getLogger(__name__)

MAX_INSTANCE_NAME_LENGTH = 63

StrOrBytesPath: TypeAlias = str | bytes | PathLike[str] | PathLike[bytes]


class Executor(ABC):
    """Interfaces to execute commands and move data in/out of an environment."""

    @abstractmethod
    def execute_popen(
        self,
        command: list[str],
        *,
        cwd: pathlib.PurePath | None = None,
        env: dict[str, str | None] | None = None,
        timeout: float | None = None,
        **kwargs: Any,
    ) -> subprocess.Popen[str]:
        """Execute a command in instance, using subprocess.Popen().

        The process' environment will inherit the execution environment's
        default environment (PATH, etc.), but can be additionally configured via
        env parameter.

        :param command: Command to execute.
        :param cwd: Working directory for the process inside the instance.
        :param env: Additional environment to set for process.
        :param timeout: Timeout (in seconds) for the command.
        :param kwargs: Additional keyword arguments to pass.

        :returns: Popen instance.
        """

    # This is modified from typeshed. The actual implementation doesn't have all this,
    # but that's just because it passes it through to subprocess.run.
    # https://github.com/python/typeshed/blob/cb2c371676f8f4a6a85b0d65c672dae308f51ca6/stdlib/subprocess.pyi#L298
    @overload
    def execute_run(
        self,
        command: list[str],
        *,
        bufsize: int = -1,
        executable: StrOrBytesPath | None = None,
        stdin: None | int | IO[Any] = None,
        stdout: None | int | IO[Any] = None,
        stderr: None | int | IO[Any] = None,
        preexec_fn: Callable[[], Any] | None = None,
        close_fds: bool = True,
        shell: bool = False,
        cwd: pathlib.PurePath | None = None,
        env: dict[str, str | None] | None = None,
        universal_newlines: bool | None = None,
        creationflags: int = 0,
        restore_signals: bool = True,
        start_new_session: bool = False,
        pass_fds: Collection[int] = (),
        capture_output: bool = False,
        check: bool = False,
        encoding: str | None = None,
        errors: str | None = None,
        input: str | None = None,
        text: Literal[True],
        timeout: float | None = None,
        user: str | int | None = None,
        group: str | int | None = None,
        extra_groups: Iterable[str | int] | None = None,
        umask: int = -1,
        pipesize: int = -1,
        **kwargs: Any,
    ) -> subprocess.CompletedProcess[str]: ...
    @overload
    def execute_run(
        self,
        command: list[str],
        *,
        bufsize: int = -1,
        executable: StrOrBytesPath | None = None,
        stdin: None | int | IO[Any] = None,
        stdout: None | int | IO[Any] = None,
        stderr: None | int | IO[Any] = None,
        preexec_fn: Callable[[], Any] | None = None,
        close_fds: bool = True,
        shell: bool = False,
        cwd: pathlib.PurePath | None = None,
        env: dict[str, str | None] | None = None,
        universal_newlines: bool | None = None,
        creationflags: int = 0,
        restore_signals: bool = True,
        start_new_session: bool = False,
        pass_fds: Collection[int] = (),
        capture_output: bool = False,
        check: bool = False,
        encoding: str,
        errors: str | None = None,
        input: str | None = None,
        text: bool | None = None,
        timeout: float | None = None,
        user: str | int | None = None,
        group: str | int | None = None,
        extra_groups: Iterable[str | int] | None = None,
        umask: int = -1,
        pipesize: int = -1,
        **kwargs: Any,
    ) -> subprocess.CompletedProcess[str]: ...
    @overload
    def execute_run(
        self,
        command: list[str],
        *,
        bufsize: int = -1,
        executable: StrOrBytesPath | None = None,
        stdin: None | int | IO[Any] = None,
        stdout: None | int | IO[Any] = None,
        stderr: None | int | IO[Any] = None,
        preexec_fn: Callable[[], Any] | None = None,
        close_fds: bool = True,
        shell: bool = False,
        cwd: pathlib.PurePath | None = None,
        env: dict[str, str | None] | None = None,
        universal_newlines: bool | None = None,
        creationflags: int = 0,
        restore_signals: bool = True,
        start_new_session: bool = False,
        pass_fds: Collection[int] = (),
        capture_output: bool = False,
        check: bool = False,
        encoding: str | None = None,
        errors: str,
        input: str | None = None,
        text: bool | None = None,
        timeout: float | None = None,
        user: str | int | None = None,
        group: str | int | None = None,
        extra_groups: Iterable[str | int] | None = None,
        umask: int = -1,
        pipesize: int = -1,
        **kwargs: Any,
    ) -> subprocess.CompletedProcess[str]: ...
    @overload
    def execute_run(
        self,
        command: list[str],
        *,
        bufsize: int = -1,
        executable: StrOrBytesPath | None = None,
        stdin: None | int | IO[Any] = None,
        stdout: None | int | IO[Any] = None,
        stderr: None | int | IO[Any] = None,
        preexec_fn: Callable[[], Any] | None = None,
        close_fds: bool = True,
        shell: bool = False,
        cwd: pathlib.PurePath | None = None,
        env: dict[str, str | None] | None = None,
        universal_newlines: Literal[True],
        creationflags: int = 0,
        restore_signals: bool = True,
        start_new_session: bool = False,
        pass_fds: Collection[int] = (),
        capture_output: bool = False,
        check: bool = False,
        encoding: str | None = None,
        errors: str | None = None,
        input: str | None = None,
        text: bool | None = None,
        timeout: float | None = None,
        user: str | int | None = None,
        group: str | int | None = None,
        extra_groups: Iterable[str | int] | None = None,
        umask: int = -1,
        pipesize: int = -1,
        **kwargs: Any,
    ) -> subprocess.CompletedProcess[str]: ...
    @overload
    def execute_run(
        self,
        command: list[str],
        *,
        bufsize: int = -1,
        executable: StrOrBytesPath | None = None,
        stdin: None | int | IO[Any] = None,
        stdout: None | int | IO[Any] = None,
        stderr: None | int | IO[Any] = None,
        preexec_fn: Callable[[], Any] | None = None,
        close_fds: bool = True,
        shell: bool = False,
        cwd: pathlib.PurePath | None = None,
        env: dict[str, str | None] | None = None,
        universal_newlines: Literal[False] | None = None,
        creationflags: int = 0,
        restore_signals: bool = True,
        start_new_session: bool = False,
        pass_fds: Collection[int] = (),
        capture_output: bool = False,
        check: bool = False,
        encoding: None = None,
        errors: None = None,
        input: Buffer | None = None,
        text: Literal[False] | None = None,
        timeout: float | None = None,
        user: str | int | None = None,
        group: str | int | None = None,
        extra_groups: Iterable[str | int] | None = None,
        umask: int = -1,
        pipesize: int = -1,
        **kwargs: Any,
    ) -> subprocess.CompletedProcess[bytes]: ...
    @overload
    def execute_run(
        self,
        command: list[str],
        *,
        bufsize: int = -1,
        executable: StrOrBytesPath | None = None,
        stdin: None | int | IO[Any] = None,
        stdout: None | int | IO[Any] = None,
        stderr: None | int | IO[Any] = None,
        preexec_fn: Callable[[], Any] | None = None,
        close_fds: bool = True,
        shell: bool = False,
        cwd: pathlib.PurePath | None = None,
        env: dict[str, str | None] | None = None,
        universal_newlines: bool | None = None,
        creationflags: int = 0,
        restore_signals: bool = True,
        start_new_session: bool = False,
        pass_fds: Collection[int] = (),
        capture_output: bool = False,
        check: bool = False,
        encoding: str | None = None,
        errors: str | None = None,
        input: Buffer | str | None = None,
        text: bool | None = None,
        timeout: float | None = None,
        user: str | int | None = None,
        group: str | int | None = None,
        extra_groups: Iterable[str | int] | None = None,
        umask: int = -1,
        pipesize: int = -1,
        **kwargs: Any,
    ) -> subprocess.CompletedProcess[Any]: ...
    @abstractmethod
    def execute_run(
        self,
        command: list[str],
        *,
        cwd: pathlib.PurePath | None = None,
        env: dict[str, str | None] | None = None,
        timeout: float | None = None,
        text: bool | None = None,
        **kwargs: Any,
    ) -> subprocess.CompletedProcess[Any]:
        """Execute a command using subprocess.run().

        The process' environment will inherit the execution environment's
        default environment (PATH, etc.), but can be additionally configured via
        env parameter.

        :param command: Command to execute.
        :param cwd: Working directory for the process inside the instance.
        :param env: Additional environment to set for process.
        :param timeout: Timeout (in seconds) for the command.
        :param check: Raise an exception if the command fails.
        :param kwargs: Keyword args to pass to subprocess.run().

        :returns: Completed process.

        :raises subprocess.CalledProcessError: if command fails and check is True.
        """

    @abstractmethod
    def pull_file(self, *, source: pathlib.PurePath, destination: pathlib.Path) -> None:
        """Copy a file from the environment to host.

        :param source: Environment file to copy.
        :param destination: Host file path to copy to.  Parent directory
            (destination.parent) must exist.

        :raises FileNotFoundError: If source file or destination's parent
            directory does not exist.
        :raises ProviderError: On error copying file.
        """

    @overload
    @contextlib.contextmanager
    def temporarily_pull_file(
        self, *, source: pathlib.PurePath, missing_ok: Literal[False] = False
    ) -> Generator[pathlib.Path, None, None]: ...
    @overload
    @contextlib.contextmanager
    def temporarily_pull_file(
        self, *, source: pathlib.PurePath, missing_ok: Literal[True]
    ) -> Generator[pathlib.Path | None, None, None]: ...
    @contextlib.contextmanager
    def temporarily_pull_file(
        self, *, source: pathlib.PurePath, missing_ok: bool = False
    ) -> Generator[pathlib.Path | None, None, None]:
        """Copy a file from the environment to a temporary file in the host.

        This is mainly a layer above `pull_file` that pulls the file into a
        temporary path which is cleaned later.

        Works as a context manager, provides the file path in the host as target.

        The temporary file is stored in the home directory where Multipass has access.

        :param source: Environment file to copy.
        :param missing_ok: Do not raise an error if the file does not exist in the
            environment; in this case the target will be None.

        :raises FileNotFoundError: If source file or destination's parent
            directory does not exist (and `missing_ok` is False).
        :raises ProviderError: On error copying file content.
        """
        with craft_providers.util.temp_paths.home_temporary_file() as tmp_file:
            try:
                self.pull_file(source=source, destination=tmp_file)
            except FileNotFoundError:
                if missing_ok:
                    yield None
                else:
                    raise
            else:
                yield tmp_file

    @abstractmethod
    def push_file(self, *, source: pathlib.Path, destination: pathlib.PurePath) -> None:
        """Copy a file from the host into the environment.

        The destination file is overwritten if it exists.

        :param source: Host file to copy.
        :param destination: Target environment file path to copy to.  Parent
            directory (destination.parent) must exist.

        :raises FileNotFoundError: If source file or destination's parent
            directory does not exist.
        :raises ProviderError: On error copying file.
        """

    @abstractmethod
    def push_file_io(
        self,
        *,
        destination: pathlib.PurePath,
        content: io.BytesIO,
        file_mode: str,
        group: str = "root",
        user: str = "root",
    ) -> None:
        """Create or replace a file with specified content and file mode.

        :param destination: Path to file.
        :param content: Contents of file.
        :param file_mode: File mode string (e.g. '0644').
        :param group: File owner group.
        :param user: File owner user.
        """

    @abstractmethod
    def delete(self) -> None:
        """Delete instance."""

    @abstractmethod
    def exists(self) -> bool:
        """Check if instance exists.

        :returns: True if instance exists.
        """

    @abstractmethod
    def mount(self, *, host_source: pathlib.Path, target: pathlib.PurePath) -> None:
        """Mount host source directory to target mount point."""

    @abstractmethod
    def is_running(self) -> bool:
        """Check if instance is running.

        :returns: True if instance is running.
        """


def get_instance_name(name: str, error_class: type[ProviderError]) -> str:
    """Get an instance-friendly name from a name.

    LXD and Multipass instance names have the same naming convention
    as Linux hostnames.

    Naming convention:
    - between 1 and 63 characters long
    - made up exclusively of letters, numbers, and hyphens from the ASCII table
    - not begin with a digit or a hyphen
    - not end with a hyphen

    To create an instance name, invalid characters are removed, the name is
    truncated to 40 characters, then a hash is appended:
    <truncated-name>-<hash-of-name>
    └     1 - 40   ┘1└     20     ┘

    :param name: the name to convert
    :param error_class: the exception class to raise if name is invalid

    :raises error_class: if name contains no alphanumeric characters
    :returns: the instance name
    """
    # remove anything that is not an alphanumeric character or hyphen
    name_with_valid_chars = re.sub(r"[^a-zA-Z0-9-]", "", name, flags=re.ASCII)
    if not name_with_valid_chars:
        raise error_class(
            brief=f"failed to create an instance with name {name!r}.",
            details="name must contain at least one alphanumeric character",
        )

    # trim digits and hyphens from the beginning and hyphens from the end
    trimmed_name = re.compile(r"^[0-9-]*(?P<valid_name>.*?)[-]*$").search(
        name_with_valid_chars
    )
    if not trimmed_name or not trimmed_name.group("valid_name"):
        raise error_class(
            brief=f"failed to create an instance with name {name!r}.",
            details="name must contain at least one alphanumeric character",
        )
    valid_name = trimmed_name.group("valid_name")

    # if the original name satisfies the naming convention, then use the original name
    if name == valid_name and len(name) <= MAX_INSTANCE_NAME_LENGTH:
        instance_name = name

    # else, continue converting the name
    else:
        # truncate to 40 characters
        truncated_name = valid_name[:40]
        # hash the entire name, not the truncated name
        hashed_name = hashlib.sha1(name.encode()).hexdigest()[:20]  # noqa: S324, security of this does not matter
        instance_name = f"{truncated_name}-{hashed_name}"

    logger.debug("Converted name %r to instance name %r", name, instance_name)
    return instance_name



# Copyright 2023 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License version 3, as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranties of MERCHANTABILITY,
# SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.
import os
import types

import pytest


@pytest.fixture
def project_main_module() -> types.ModuleType:
    """Fixture that returns the project's principal package (imported).

    This fixture should be rewritten by "downstream" projects to return the correct
    module. Then, every test that uses this fixture will correctly test against the
    downstream project.
    """
    try:
        # This should be the project's main package; downstream projects must update this.
        import craft_providers  # noqa: PLC0415

        main_module = craft_providers
    except ImportError:
        pytest.fail(
            "Failed to import the project's main module: check if it needs updating",
        )
    return main_module


@pytest.fixture(scope="session", autouse=True)
def _suppress_upgrade() -> None:
    """Suppress dist-upgrade when starting containers during tests.

    This speeds up our tests since dist-upgrade can take a while.
    """
    os.environ["CRAFT_PROVIDERS_EXPERIMENTAL_SUPPRESS_UPGRADE_UNSUPPORTED"] = "1"



# Copyright 2023 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License version 3, as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranties of MERCHANTABILITY,
# SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.
"""Integration tests related to building the package."""

import re
import subprocess
import sys
from pathlib import Path
from zipfile import ZipFile

import pytest

pytestmark = [pytest.mark.slow]


def test_packages(project_main_module, tmp_path, request):
    """Check wheel generation from our pyproject.toml"""
    root_dir = Path(request.config.rootdir)
    out_dir = tmp_path
    subprocess.check_call(
        [sys.executable, "-m", "build", "--outdir", out_dir, root_dir],
    )
    wheels = list(tmp_path.glob("*.whl"))
    assert len(wheels) == 1
    wheel = wheels[0]

    main_module = project_main_module.__name__

    project_files = []

    dist_files = []
    dist_info_re = re.compile(f"{main_module}-.*.dist-info")

    invalid = []

    with ZipFile(wheel) as wheel_zip:
        names = [Path(p) for p in wheel_zip.namelist()]
        assert len(names) > 1
        for name in names:
            top = name.parts[0]
            if top == main_module:
                project_files.append(name)
            elif dist_info_re.match(top):
                dist_files.append(top)
            else:
                invalid = []

    # Only the top-level "project_name" dir should be present, plus the
    # project_name-xyz-dist-info/ entries.
    assert project_files, f"No '{main_module}' modules were packaged!"
    assert dist_files, "The dist-info directory was not created!"
    assert not invalid, f"Invalid files were packaged: {invalid}"



# Copyright 2025 Canonical Ltd.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
"""Integration tests to test the shutdown/relaunch workflow."""

import io
import pathlib
import time
from typing import TYPE_CHECKING, cast

import craft_providers
import pytest
from craft_providers import bases
from craft_providers.bases.almalinux import AlmaLinuxBaseAlias
from craft_providers.bases.centos import CentOSBaseAlias
from craft_providers.bases.ubuntu import BuilddBaseAlias
from craft_providers.multipass.multipass_provider import MultipassProvider

if TYPE_CHECKING:
    from craft_providers.executor import Executor


@pytest.mark.slow
@pytest.mark.parametrize(
    "base_alias",
    [
        *bases.almalinux.AlmaLinuxBaseAlias,
        bases.BuilddBaseAlias.NOBLE,
        # https://github.com/canonical/craft-providers/issues/765
        # We should enable all of these for weekly tests.
        # *bases.ubuntu.BuilddBaseAlias,
    ],
)
def test_relaunch(
    session_provider: craft_providers.Provider,
    base_alias: BuilddBaseAlias | CentOSBaseAlias | AlmaLinuxBaseAlias,
    tmp_path: pathlib.Path,
):
    if (
        isinstance(session_provider, MultipassProvider)
        and base_alias not in bases.ubuntu.BuilddBaseAlias
    ):
        pytest.skip("Non-Ubuntu bases not supported with Multipass.")
    if base_alias == bases.ubuntu.BuilddBaseAlias.XENIAL:
        pytest.skip(
            "Xenial not supported: https://github.com/canonical/craft-providers/issues/582"
        )
    if base_alias == bases.ubuntu.BuilddBaseAlias.ORACULAR:
        pytest.skip(
            "Oracular is unsupported: https://github.com/canonical/craft-providers/issues/598"
        )

    base_cls = bases.get_base_from_alias(base_alias)
    base = base_cls(alias=base_alias)  # type: ignore[reportArgumentType, arg-type]
    project_name = f"relaunch-{base_alias.name}"

    try:
        # Set up both a file that should exist for the whole run and one that should
        # disappear after a proper shutdown.
        with session_provider.launched_environment(
            project_name=project_name,
            project_path=tmp_path,
            base_configuration=base,
            instance_name="relaunch-test",
            allow_unstable=True,
            shutdown_delay_mins=1,
        ) as instance:
            instance.execute_run(["touch", "/tmp/session-file"], check=True)
            instance.execute_run(["touch", "/root/permanent-file"], check=True)

            # Alma Linux only clears tmp files after 10 days by default.
            # This configures systemd-tmpfiles to clear them on every boot.
            if isinstance(base_alias, AlmaLinuxBaseAlias):
                content = io.BytesIO(b"r! /tmp/* 1777 root root 0")
                instance.push_file_io(
                    destination=pathlib.PurePosixPath(
                        "/etc/tmpfiles.d/craft-tempfiles.conf"
                    ),
                    content=content,
                    file_mode="644",
                )

        assert instance.is_running()

        # Check that both files still exist after a delayed shutdown
        with session_provider.launched_environment(
            project_name=project_name,
            project_path=tmp_path,
            base_configuration=base,
            instance_name="relaunch-test",
            allow_unstable=True,
            shutdown_delay_mins=0,
        ) as instance:
            instance.execute_run(["ls", "/tmp/session-file"], check=True)
            instance.execute_run(["ls", "/root/permanent-file"], check=True)
            # Check that the shutdown was properly cancelled.
            result = instance.execute_run(["shutdown", "--show"], check=False)
            assert result.returncode == 1

        # Instance will take a little while to shut down asynchronously. Wait until
        # it's no longer running before moving on.
        for _ in range(100):
            if instance.is_running():
                time.sleep(0.1)
        assert not instance.is_running()

        # Check that the file in /tmp got deleted after a real shutdown, but that the
        # permanent file is still there.
        with session_provider.launched_environment(
            project_name="craft-providers-integration-relaunch",
            project_path=tmp_path,
            base_configuration=base,
            instance_name="relaunch-test",
            allow_unstable=True,
            shutdown_delay_mins=None,
        ) as instance:
            result = instance.execute_run(["ls", "/tmp/session-file"], check=False)
            assert result.returncode == 2
            instance.execute_run(["ls", "/root/permanent-file"], check=True)

    finally:
        instance = cast("Executor", locals().get("instance"))
        if instance is not None:
            instance.delete()



#
# Copyright 2021 Canonical Ltd.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#

import pytest
import responses
from craft_providers.lxd import LXD, LXDError


def test_init(fake_process):
    fake_process.register_subprocess(
        [
            "lxd",
            "init",
        ]
    )

    LXD().init()

    assert len(fake_process.calls) == 1


def test_init_all_opts(fake_process):
    fake_process.register_subprocess(
        [
            "sudo",
            "lxd",
            "init",
            "--auto",
        ]
    )

    LXD().init(
        auto=True,
        sudo=True,
    )

    assert len(fake_process.calls) == 1


def test_init_error(fake_process):
    fake_process.register_subprocess(
        [
            "lxd",
            "init",
        ],
        returncode=1,
    )

    with pytest.raises(LXDError, match="Failed to init LXD."):
        LXD().init()


@pytest.mark.parametrize(
    ("version", "compatible"),
    [
        ("3.0", False),
        ("3.1.4", False),
        ("3.10", False),
        ("4.0", True),
        ("4.1.4", True),
        ("4.10", True),
        ("5.21.0", True),
        ("6.5", True),
    ],
)
def test_is_supported_version(mocker, version, compatible):
    lxd = LXD()

    mocker.patch.object(lxd, "version", return_value=version)

    assert lxd.is_supported_version() == compatible


@pytest.mark.parametrize("version_data", ["", "invalid"])
def test_is_supported_version_parse_error_assumes_true(mocker, version_data):
    lxd = LXD()

    mocker.patch.object(lxd, "version", return_value=version_data)

    assert lxd.is_supported_version()


@responses.activate
def test_version_from_socket(fake_process):  # Error if using subprocess.
    # Snap socket
    responses.add(
        responses.GET,
        "http+unix://%2Fvar%2Fsnap%2Flxd%2Fcommon%2Flxd%2Funix.socket/1.0",
        json={
            "type": "sync",
            "metadata": {"environment": {"server_version": "test-version"}},
        },
        status=200,
    )
    # If installed from apt or similar.
    responses.add(
        responses.GET,
        "http+unix://%2Fvar%2Flib%2Flxd%2Funix.socket/1.0",
        json={
            "type": "sync",
            "metadata": {"environment": {"server_version": "test-version"}},
        },
        status=200,
    )

    version = LXD().version()

    assert version == "test-version"


@responses.activate  # Simply error on any request to the LXD socket.
def test_version_from_command(fake_process):
    fake_process.register_subprocess(
        [
            "lxd",
            "version",
        ],
        stdout="test-version",
    )

    version = LXD().version()

    assert len(fake_process.calls) == 1
    assert version == "test-version"


@responses.activate  # Simply error on any request to the LXD socket.
def test_version_error(fake_process):
    fake_process.register_subprocess(
        [
            "lxd",
            "version",
        ],
        returncode=1,
    )

    with pytest.raises(LXDError, match="Failed to query LXD version."):
        LXD().version()


def test_wait_ready(fake_process):
    fake_process.register_subprocess(
        [
            "lxd",
            "waitready",
        ]
    )

    LXD().wait_ready()

    assert len(fake_process.calls) == 1


def test_wait_ready_all_opts(fake_process):
    fake_process.register_subprocess(
        [
            "sudo",
            "lxd",
            "waitready",
            "--timeout=5",
        ]
    )

    LXD().wait_ready(
        sudo=True,
        timeout=5,
    )

    assert len(fake_process.calls) == 1


def test_wait_ready_error(fake_process):
    fake_process.register_subprocess(
        [
            "lxd",
            "waitready",
        ],
        returncode=1,
    )

    with pytest.raises(LXDError, match="Failed to wait for LXD to get ready."):
        LXD().wait_ready()



#
# Copyright 2021-2025 Canonical Ltd.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#

"""Base compatibility checks."""

from __future__ import annotations

import logging
import platform
from typing import TYPE_CHECKING, TypedDict

from craft_providers.bases.ubuntu import BuilddBase, BuilddBaseAlias
from craft_providers.errors import (
    ProviderError,
)
from craft_providers.util.os_release import parse_os_release

if TYPE_CHECKING:
    from enum import Enum

    from craft_providers.base import Base
    from craft_providers.executor import Executor

logger = logging.getLogger(__name__)


class InvalidVersionSet(TypedDict):
    """A set of invalid version combinations for runtime assertions."""

    host_less_than_equal: BuilddBaseAlias
    guest_greater_than_equal: BuilddBaseAlias
    lxd_less_than: list[tuple[int, int, int]]
    kernel_less_than: tuple[int, int]


INVALID_VERSIONS: list[InvalidVersionSet] = [
    {
        "host_less_than_equal": BuilddBaseAlias.FOCAL,
        "guest_greater_than_equal": BuilddBaseAlias.ORACULAR,
        # The system is affected by the cgroups bug if both of the above and either of the below
        "lxd_less_than": [
            (5, 0, 4),
            (5, 21, 2),
        ],
        "kernel_less_than": (5, 15),
    },
]


def _lxd_version_match(
    system_version: tuple[int, int, int],
    affected_versions: list[tuple[int, int, int]],
) -> bool:
    """Compare the system lxd version with the list of affected versions.

    :returns: True if the system lxd version is affected, False if it is not, or if
    whether it is affected can't be determined.
    """
    # First look for matching major/minor, and compare patch
    for affected_version in affected_versions:
        if (
            affected_version[0] == system_version[0]
            and affected_version[1] == system_version[1]
        ):
            return system_version[2] < affected_version[2]

    # Assume major versions below those listed are affected, otherwise either not
    # affected or we can't tell so we won't fail.
    lowest_major = min([v[0] for v in affected_versions])
    return system_version[0] < lowest_major


def ensure_guest_compatible(
    base_configuration: Base[Enum],
    instance: Executor,
    lxd_version: str,
) -> None:
    """Ensure host is compatible with guest instance."""
    if not issubclass(type(base_configuration), BuilddBase):
        # Not ubuntu, not sure how to check
        logger.debug(
            f"Base alias configuration is {base_configuration.alias!r}: no checks for non Buildd"
        )
        return

    host_os_release = parse_os_release()
    # Return early for non Ubuntu hosts
    if host_os_release.get("ID") != "ubuntu":
        logger.debug(
            f"Host is {host_os_release.get('ID')}: no checks for non Ubuntu hosts"
        )
        return

    host_base_alias = BuilddBaseAlias(host_os_release.get("VERSION_ID"))

    guest_os_release = base_configuration.get_os_release(executor=instance)
    guest_base_alias = BuilddBaseAlias(guest_os_release.get("VERSION_ID"))

    # Strip off anything after the first space - sometimes "LTS" is appended
    lxd_version_split = lxd_version.strip().split(" ")[0].split(".")
    lxd_major = int(lxd_version_split[0])
    lxd_minor = int(lxd_version_split[1])
    try:
        lxd_patch = int(lxd_version_split[2])
    except IndexError:
        # LXD version strings sometimes omit the patch - call it zero
        lxd_patch = 0
    lxd_version_tup = (lxd_major, lxd_minor, lxd_patch)

    kernel_version_tup = tuple([int(v) for v in platform.release().split(".")[0:2]])

    # If the host OS is focal (20.04) or older, and the guest OS is oracular (24.10)
    # or newer, then the host lxd must be >=5.0.4 or >=5.21.2, and kernel must be
    # 5.15 or newer.  Otherwise, weird systemd failures will occur due to a mismatch
    # between cgroupv1 and v2 support.
    # https://discourse.ubuntu.com/t/lxd-5-0-4-lts-has-been-released/49681#p-123331-support-for-ubuntu-oracular-containers-on-cgroupv2-hosts

    for invalid in INVALID_VERSIONS:
        if (
            host_base_alias <= invalid["host_less_than_equal"]
            and guest_base_alias >= invalid["guest_greater_than_equal"]
            and (
                _lxd_version_match(
                    lxd_version_tup,
                    invalid["lxd_less_than"],
                )
                or kernel_version_tup < invalid["kernel_less_than"]
            )
        ):
            raise ProviderError(
                brief="This combination of guest and host OS versions requires a newer kernel and/or lxd.",
                resolution="Ensure you have lxd>=5.21.2 or >= 5.0.4, and kernel>=5.15 - try the lxd snap or HWE kernel.",
            )



#
# Copyright 2021-2023 Canonical Ltd.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#

"""Unit tests for Ubuntu bases."""

import importlib.resources
import pathlib
import re
import subprocess
import tempfile
from pathlib import Path
from textwrap import dedent
from unittest.mock import ANY, call, patch

import pytest
import requests
from craft_providers.actions.snap_installer import Snap, SnapInstallationError
from craft_providers.bases import ubuntu
from craft_providers.errors import (
    BaseCompatibilityError,
    BaseConfigurationError,
    NetworkError,
)
from craft_providers.instance_config import InstanceConfiguration
from freezegun import freeze_time
from logassert import Exact  # type: ignore[import-untyped]

from tests.unit.conftest import DEFAULT_FAKE_CMD

TEST_DIR = pathlib.Path(__file__).parent.parent / "data"


@pytest.fixture
def mock_load(mocker):
    return mocker.patch(
        "craft_providers.instance_config.InstanceConfiguration.load",
        return_value=InstanceConfiguration(compatibility_tag="buildd-base-v7"),
    )


@pytest.fixture
def fake_filesystem(fs):
    return fs


@pytest.fixture
def mock_install_from_store(mocker):
    return mocker.patch("craft_providers.actions.snap_installer.install_from_store")


@pytest.fixture
def mock_inject_from_host(mocker):
    return mocker.patch("craft_providers.actions.snap_installer.inject_from_host")


@pytest.fixture
def mock_get_os_release(mocker):
    return mocker.patch.object(
        ubuntu.BuilddBase,
        "get_os_release",
        return_value={
            "NAME": "Ubuntu",
            "VERSION_ID": "22.04",
            "VERSION_CODENAME": "jammy",
        },
    )


@pytest.fixture
def mock_requests_head(mocker, request):
    status_code = getattr(request, "param", 404)
    mock_response = mocker.Mock(status_code=status_code)
    return mocker.patch("requests.head", return_value=mock_response)


@pytest.fixture
def fake_update_commands(fake_process):
    """Allow running update commands with a fake process."""
    fake_process.register_subprocess([*DEFAULT_FAKE_CMD, "apt-get", "update"])
    fake_process.register_subprocess(
        [*DEFAULT_FAKE_CMD, "apt-get", "-y", "dist-upgrade"]
    )


@pytest.mark.usefixtures("fake_update_commands")
@pytest.mark.parametrize("alias", list(ubuntu.BuilddBaseAlias))
@pytest.mark.parametrize(
    ("environment", "etc_environment_content"),
    [
        (
            None,
            (
                b"PATH=/usr/local/sbin:/usr/local/bin:"
                b"/usr/sbin:/usr/bin:/sbin:/bin:/snap/bin\n"
                b"DEBIAN_FRONTEND=noninteractive\n"
                b"DEBCONF_NONINTERACTIVE_SEEN=true\n"
                b"DEBIAN_PRIORITY=critical\n"
            ),
        ),
        (
            {
                "https_proxy": "http://foo.bar:8081",
                "PATH": "/snap",
                "http_proxy": "http://foo.bar:8080",
            },
            (
                b"https_proxy=http://foo.bar:8081\n"
                b"PATH=/snap\nhttp_proxy=http://foo.bar:8080\n"
                b"DEBIAN_FRONTEND=noninteractive\n"
                b"DEBCONF_NONINTERACTIVE_SEEN=true\n"
                b"DEBIAN_PRIORITY=critical\n"
            ),
        ),
    ],
)
@pytest.mark.parametrize("no_cdn", [False, True])
@pytest.mark.parametrize(
    ("snaps", "expected_snap_call"),
    [
        (None, []),
        (
            [Snap(name="snap1", channel="edge", classic=True)],
            [call(executor=ANY, snap_name="snap1", channel="edge", classic=True)],
        ),
    ],
)
@pytest.mark.parametrize(
    ("packages", "expected_packages"),
    [
        (
            None,
            [
                "apt-utils",
                "build-essential",
                "curl",
                "fuse",
                "udev",
                "python3",
                "python3-dev",
                "python3-pip",
                "python3-wheel",
                "python3-setuptools",
            ],
        ),
        (
            ["grep", "git"],
            [
                "apt-utils",
                "build-essential",
                "curl",
                "fuse",
                "udev",
                "python3",
                "python3-dev",
                "python3-pip",
                "python3-wheel",
                "python3-setuptools",
                "grep",
                "git",
            ],
        ),
    ],
)
@pytest.mark.parametrize(
    ("tag", "expected_tag"), [(None, "buildd-base-v7"), ("test-tag", "test-tag")]
)
def test_setup(
    fake_process,
    fake_executor,
    fake_filesystem,
    alias,
    environment,
    etc_environment_content,
    no_cdn,
    mock_load,
    mock_inject_from_host,
    mock_install_from_store,
    mock_requests_head,
    mocker,
    snaps,
    expected_snap_call,
    packages,
    expected_packages,
    tag,
    expected_tag,
):
    with importlib.resources.path(
        "craft_providers.data", "ubuntu.csv"
    ) as distro_info_file:
        fake_filesystem.add_real_file(distro_info_file)

    mock_load.return_value = InstanceConfiguration(compatibility_tag=expected_tag)

    if environment is None:
        environment = ubuntu.BuilddBase.default_command_environment()

    if no_cdn:
        fake_filesystem.create_file(
            "/etc/systemd/system/snapd.service.d/no-cdn.conf",
            contents=dedent(
                """\
                [Service]
                Environment=SNAPPY_STORE_NO_CDN=1
                """
            ),
        )

    base_config = ubuntu.BuilddBase(
        alias=alias,
        compatibility_tag=tag,
        environment=environment,
        hostname="test-hostname",
        snaps=snaps,
        packages=packages,
    )

    fake_process.register_subprocess(
        [*DEFAULT_FAKE_CMD, "cat", "/etc/os-release"],
        stdout=dedent(
            f"""\
            NAME="Ubuntu"
            ID=ubuntu
            ID_LIKE=debian
            VERSION_ID="{alias.value}"
            """
        ),
    )
    fake_process.register_subprocess(
        [*DEFAULT_FAKE_CMD, "test", "-f", "/etc/craft-instance.conf"],
        returncode=1,
    )
    fake_process.register_subprocess(
        [*DEFAULT_FAKE_CMD, "systemctl", "is-system-running"], stdout="degraded"
    )
    fake_process.register_subprocess(
        [*DEFAULT_FAKE_CMD, "hostname", "-F", "/etc/hostname"]
    )
    fake_process.register_subprocess([*DEFAULT_FAKE_CMD, "chmod", "go+x", "/root"])
    fake_process.register_subprocess(
        [
            *DEFAULT_FAKE_CMD,
            "ln",
            "-sf",
            "/run/systemd/resolve/resolv.conf",
            "/etc/resolv.conf",
        ]
    )
    fake_process.register_subprocess(
        [*DEFAULT_FAKE_CMD, "systemctl", "enable", "systemd-resolved"]
    )
    fake_process.register_subprocess(
        [*DEFAULT_FAKE_CMD, "systemctl", "restart", "systemd-resolved"]
    )
    fake_process.register_subprocess(
        [*DEFAULT_FAKE_CMD, "systemctl", "enable", "systemd-networkd"]
    )
    fake_process.register_subprocess(
        [*DEFAULT_FAKE_CMD, "systemctl", "restart", "systemd-networkd"]
    )
    fake_process.register_subprocess(
        [*DEFAULT_FAKE_CMD, "getent", "hosts", "snapcraft.io"]
    )
    fake_process.register_subprocess(
        [*DEFAULT_FAKE_CMD, "cat", "/etc/os-release"],
        stdout=dedent(
            f"""\
            NAME="Ubuntu"
            ID=ubuntu
            ID_LIKE=debian
            VERSION_ID="{alias.value}"
            VERSION_CODENAME="test-name"
            UBUNTU_CODENAME="noble"
            """
        ),
    )
    fake_process.register_subprocess(
        [*DEFAULT_FAKE_CMD, "apt-get", "install", "-y", *expected_packages]
    )
    fake_process.register_subprocess(
        [*DEFAULT_FAKE_CMD, "apt-get", "install", "-y", "fuse", "udev"]
    )
    fake_process.register_subprocess([*DEFAULT_FAKE_CMD, "apt-get", "autoremove", "-y"])
    fake_process.register_subprocess([*DEFAULT_FAKE_CMD, "apt-get", "clean", "-y"])
    fake_process.register_subprocess(
        [*DEFAULT_FAKE_CMD, "systemctl", "is-active", "systemd-udevd"],
        stdout="active",
    )
    fake_process.register_subprocess(
        [*DEFAULT_FAKE_CMD, "systemctl", "enable", "systemd-udevd"]
    )
    fake_process.register_subprocess(
        [*DEFAULT_FAKE_CMD, "systemctl", "start", "systemd-udevd"]
    )
    if no_cdn:
        fake_process.register_subprocess(
            [*DEFAULT_FAKE_CMD, "mkdir", "-p", "/etc/systemd/system/snapd.service.d"]
        )
    fake_process.register_subprocess(
        [*DEFAULT_FAKE_CMD, "apt-get", "install", "-y", "snapd"]
    )
    fake_process.register_subprocess(
        [*DEFAULT_FAKE_CMD, "systemctl", "start", "snapd.socket"]
    )
    fake_process.register_subprocess(
        [*DEFAULT_FAKE_CMD, "systemctl", "restart", "snapd.service"]
    )
    fake_process.register_subprocess(
        [*DEFAULT_FAKE_CMD, "snap", "wait", "system", "seed.loaded"]
    )
    fake_process.register_subprocess([*DEFAULT_FAKE_CMD, "snap", "refresh", "--hold"])
    fake_process.register_subprocess(
        [*DEFAULT_FAKE_CMD, "snap", "watch", "--last=auto-refresh?"]
    )
    fake_process.register_subprocess(
        [*DEFAULT_FAKE_CMD, "snap", "set", "system", "proxy.http=http://foo.bar:8080"]
    )
    fake_process.register_subprocess(
        [*DEFAULT_FAKE_CMD, "snap", "unset", "system", "proxy.http"]
    )
    fake_process.register_subprocess(
        [*DEFAULT_FAKE_CMD, "snap", "set", "system", "proxy.https=http://foo.bar:8081"]
    )
    fake_process.register_subprocess(
        [*DEFAULT_FAKE_CMD, "snap", "unset", "system", "proxy.https"]
    )
    fake_process.register_subprocess(
        [*DEFAULT_FAKE_CMD, "ln", "-sf", "/var/lib/snapd/snap", "/snap"]
    )
    fake_process.register_subprocess(
        [*DEFAULT_FAKE_CMD, "systemctl", "enable", "--now", "snapd.socket"]
    )

    base_config.setup(executor=fake_executor)

    expected_push_file_io = [
        {
            "destination": "/etc/craft-instance.conf",
            "content": (f"compatibility_tag: {expected_tag}\nsetup: false\n").encode(),
            "file_mode": "0644",
            "group": "root",
            "user": "root",
        },
        {
            "destination": "/etc/craft-instance.conf",
            "content": (f"compatibility_tag: {expected_tag}\n").encode(),
            "file_mode": "0644",
            "group": "root",
            "user": "root",
        },
        {
            "destination": "/etc/environment",
            "content": etc_environment_content,
            "file_mode": "0644",
            "group": "root",
            "user": "root",
        },
        {
            "destination": "/etc/apt/apt.conf.d/20auto-upgrades",
            "content": dedent(
                """\
                APT::Periodic::Update-Package-Lists "10000";
                APT::Periodic::Unattended-Upgrade "0";
                """
            ).encode(),
            "file_mode": "0644",
            "group": "root",
            "user": "root",
        },
        {
            "destination": "/etc/hostname",
            "content": b"test-hostname\n",
            "file_mode": "0644",
            "group": "root",
            "user": "root",
        },
        {
            "destination": "/etc/systemd/network/10-eth0.network",
            "content": dedent(
                """\
                [Match]
                Name=eth0

                [Network]
                DHCP=ipv4
                LinkLocalAddressing=ipv6

                [DHCP]
                RouteMetric=100
                UseMTU=true
                """
            ).encode(),
            "file_mode": "0644",
            "group": "root",
            "user": "root",
        },
        {
            "destination": "/etc/apt/apt.conf.d/00no-recommends",
            "content": b'APT::Install-Recommends "false";\n',
            "file_mode": "0644",
            "group": "root",
            "user": "root",
        },
        {
            "destination": "/etc/apt/apt.conf.d/00update-errors",
            "content": b'APT::Update::Error-Mode "any";\n',
            "file_mode": "0644",
            "group": "root",
            "user": "root",
        },
        {
            "destination": "/etc/craft-instance.conf",
            "content": (f"compatibility_tag: {expected_tag}\nsetup: true\n").encode(),
            "file_mode": "0644",
            "group": "root",
            "user": "root",
        },
    ]
    expected_push_file = []
    if no_cdn:
        expected_push_file.append(
            {
                "source": Path("/etc/systemd/system/snapd.service.d/no-cdn.conf"),
                "destination": Path("/etc/systemd/system/snapd.service.d/no-cdn.conf"),
            }
        )

    assert fake_executor.records_of_push_file_io == expected_push_file_io
    assert fake_executor.records_of_pull_file == []
    assert fake_executor.records_of_push_file == expected_push_file
    assert mock_install_from_store.mock_calls == expected_snap_call


def test_install_snaps_install_from_store(fake_executor, mock_install_from_store):
    """Verify installing snaps calls install_from_store()."""
    my_snaps = [
        Snap(name="snap1"),
        Snap(name="snap2", channel="edge"),
        Snap(name="snap3", channel="edge", classic=True),
    ]
    base = ubuntu.BuilddBase(alias=ubuntu.BuilddBaseAlias.JAMMY, snaps=my_snaps)

    base._install_snaps(executor=fake_executor)

    assert mock_install_from_store.mock_calls == [
        call(
            executor=fake_executor, snap_name="snap1", channel="stable", classic=False
        ),
        call(executor=fake_executor, snap_name="snap2", channel="edge", classic=False),
        call(executor=fake_executor, snap_name="snap3", channel="edge", classic=True),
    ]


def test_install_snaps_inject_from_host_valid(
    fake_executor, mock_inject_from_host, mocker
):
    """Verify installing snaps calls inject_from_host()."""
    mocker.patch("sys.platform", "linux")
    my_snaps = [
        Snap(name="snap1", channel=None),
        Snap(name="snap2", channel=None, classic=True),
    ]
    base = ubuntu.BuilddBase(alias=ubuntu.BuilddBaseAlias.JAMMY, snaps=my_snaps)

    base._install_snaps(executor=fake_executor)

    assert mock_inject_from_host.mock_calls == [
        call(executor=fake_executor, snap_name="snap1", classic=False),
        call(executor=fake_executor, snap_name="snap2", classic=True),
    ]


def test_install_snaps_inject_from_host_not_linux_error(fake_executor, mocker):
    """Verify install_snaps raises an error when injecting from host on
    a non-linux system."""
    mocker.patch("sys.platform", return_value="darwin")
    my_snaps = [Snap(name="snap1", channel=None)]
    base = ubuntu.BuilddBase(alias=ubuntu.BuilddBaseAlias.JAMMY, snaps=my_snaps)

    with pytest.raises(BaseConfigurationError) as exc_info:
        base._install_snaps(executor=fake_executor)

    assert exc_info.value == BaseConfigurationError(
        brief="cannot inject snap 'snap1' from host on a non-linux system",
        resolution="install the snap from the store by setting the 'channel' parameter",
    )


def test_install_snaps_install_from_store_error(fake_executor, mocker):
    """Verify install_snaps raises an error when install_from_store fails."""
    mocker.patch(
        "craft_providers.actions.snap_installer.install_from_store",
        side_effect=SnapInstallationError(brief="test error"),
    )
    my_snaps = [Snap(name="snap1", channel="candidate")]
    base = ubuntu.BuilddBase(alias=ubuntu.BuilddBaseAlias.JAMMY, snaps=my_snaps)

    with pytest.raises(BaseConfigurationError) as exc_info:
        base._install_snaps(executor=fake_executor)

    assert exc_info.value == BaseConfigurationError(
        brief=(
            "failed to install snap 'snap1' from store"
            " channel 'candidate' in target environment."
        ),
        resolution="Check Snap store status at https://status.snapcraft.io",
    )


def test_install_snaps_inject_from_host_error(fake_executor, mocker):
    """Verify install_snaps raises an error when inject_from_host fails."""
    mocker.patch("sys.platform", "linux")
    mocker.patch(
        "craft_providers.actions.snap_installer.inject_from_host",
        side_effect=SnapInstallationError(brief="test error"),
    )
    my_snaps = [Snap(name="snap1", channel=None)]
    base = ubuntu.BuilddBase(alias=ubuntu.BuilddBaseAlias.JAMMY, snaps=my_snaps)

    with pytest.raises(BaseConfigurationError) as exc_info:
        base._install_snaps(executor=fake_executor)

    assert exc_info.value == BaseConfigurationError(
        brief="failed to inject host's snap 'snap1' into target environment."
    )


@pytest.mark.usefixtures("fake_update_commands")
def test_setup_apt(fake_executor, fake_process):
    """Verify packages are installed as expected."""
    packages = ["grep", "git"]
    base = ubuntu.BuilddBase(alias=ubuntu.BuilddBaseAlias.JAMMY, packages=packages)
    fake_process.register_subprocess(
        [
            *DEFAULT_FAKE_CMD,
            "apt-get",
            "install",
            "-y",
            "apt-utils",
            "build-essential",
            "curl",
            "fuse",
            "udev",
            "python3",
            "python3-dev",
            "python3-pip",
            "python3-wheel",
            "python3-setuptools",
            "grep",
            "git",
        ]
    )

    base._setup_packages(executor=fake_executor)


@pytest.mark.usefixtures("fake_update_commands")
def test_setup_apt_install_default(fake_executor, fake_process):
    """Verify only default packages are installed."""
    base = ubuntu.BuilddBase(alias=ubuntu.BuilddBaseAlias.JAMMY)

    fake_process.register_subprocess(
        [
            *DEFAULT_FAKE_CMD,
            "apt-get",
            "install",
            "-y",
            "apt-utils",
            "build-essential",
            "curl",
            "fuse",
            "udev",
            "python3",
            "python3-dev",
            "python3-pip",
            "python3-wheel",
            "python3-setuptools",
        ]
    )

    base._setup_packages(executor=fake_executor)


@pytest.mark.usefixtures("fake_update_commands")
def test_setup_apt_install_override_system(fake_executor, fake_process):
    """Verify override default packages."""
    base = ubuntu.BuilddBase(
        alias=ubuntu.BuilddBaseAlias.JAMMY,
        packages=["clang"],
        use_default_packages=False,
    )
    fake_process.register_subprocess(
        [
            *DEFAULT_FAKE_CMD,
            "apt-get",
            "install",
            "-y",
            "clang",
        ]
    )

    base._setup_packages(executor=fake_executor)


def test_setup_apt_install_packages_update_error(
    mocker, fake_executor, mock_requests_head, fake_process
):
    """Verify error is caught from `apt-get update` call."""

    def _raise_error(process) -> None:
        raise subprocess.CalledProcessError(100, ["error"])

    fake_process.register_subprocess(
        [*DEFAULT_FAKE_CMD, "cat", "/etc/os-release"],
        stdout="UBUNTU_CODENAME=jammy",
    )
    fake_process.register_subprocess(
        [*DEFAULT_FAKE_CMD, "apt-get", "update"],
        callback=_raise_error,
    )
    fake_process.register_subprocess(
        [*DEFAULT_FAKE_CMD, "bash", "-c", "exec 3<> /dev/tcp/snapcraft.io/443"],
    )
    base = ubuntu.BuilddBase(alias=ubuntu.BuilddBaseAlias.JAMMY)

    with pytest.raises(BaseConfigurationError) as exc_info:
        base._pre_setup_packages(executor=fake_executor)

    assert exc_info.value == BaseConfigurationError(
        brief="Failed to update apt cache.",
        details="* Command that failed: 'error'\n* Command exit code: 100",
    )


def test_setup_apt_install_packages_install_error(mocker, fake_executor, fake_process):
    """Verify error is caught from `apt-get install` call."""
    base = ubuntu.BuilddBase(alias=ubuntu.BuilddBaseAlias.JAMMY)

    fake_process.register_subprocess(
        [*DEFAULT_FAKE_CMD, "apt-get", "-y", "dist-upgrade"]
    )
    fake_process.register_subprocess(
        [
            *DEFAULT_FAKE_CMD,
            "apt-get",
            "install",
            "-y",
            fake_process.any(min=1, max=100),
        ],
        returncode=43,
    )
    fake_process.register_subprocess(
        [*DEFAULT_FAKE_CMD, "bash", "-c", "exec 3<> /dev/tcp/snapcraft.io/443"]
    )

    with pytest.raises(
        BaseConfigurationError,
        match=r"Failed to install packages.\n.+\n\* Command exit code: 43",
    ):
        base._setup_packages(executor=fake_executor)


def test_ensure_image_version_compatible_failure(fake_executor, monkeypatch):
    base_config = ubuntu.BuilddBase(alias=ubuntu.BuilddBaseAlias.JAMMY)
    monkeypatch.setattr(
        InstanceConfiguration,
        "load",
        lambda **kwargs: InstanceConfiguration(compatibility_tag="invalid-tag"),
    )

    with pytest.raises(BaseCompatibilityError) as exc_info:
        base_config._ensure_instance_config_compatible(executor=fake_executor)

    assert exc_info.value == BaseCompatibilityError(
        "Expected image compatibility tag 'buildd-base-v7', found 'invalid-tag'"
    )


@pytest.mark.parametrize("alias", list(ubuntu.BuilddBaseAlias))
@pytest.mark.parametrize(
    "cache_path", [pathlib.Path(tempfile.gettempdir(), "fake-cache-dir")]
)
def test_mount_cache_dirs(fake_process, fake_executor, cache_path: pathlib.Path, alias):
    """Test mounting of cache directories with a cache directory set."""

    base = ubuntu.BuilddBase(alias=alias, cache_path=cache_path)
    host_cache_dir = (cache_path / base.compatibility_tag / str(base.alias)).resolve()
    user_cache_dir = pathlib.Path("/root/.cache")
    fake_process.register(
        [*DEFAULT_FAKE_CMD, "bash", "-c", "echo -n ${XDG_CACHE_HOME:-${HOME}/.cache}"],
        stdout=str(user_cache_dir),
    )
    fake_process.register(
        [*DEFAULT_FAKE_CMD, "mkdir", "-p", "/root/.cache/pip"],
    )

    base._mount_shared_cache_dirs(fake_executor)

    expected_mounts = [
        {
            "host_source": host_cache_dir / "pip",
            "target": user_cache_dir / "pip",
        },
    ]
    assert fake_executor.records_of_mount == expected_mounts


def test_get_os_release(fake_process, fake_executor):
    """`get_os_release` should parse data from `/etc/os-release` to a dict."""
    base_config = ubuntu.BuilddBase(alias=ubuntu.BuilddBaseAlias.JAMMY)
    fake_process.register_subprocess(
        [*DEFAULT_FAKE_CMD, "cat", "/etc/os-release"],
        stdout="NAME=Ubuntu\nVERSION_ID=12.04\n",
    )

    result = base_config.get_os_release(executor=fake_executor)

    assert result == {"NAME": "Ubuntu", "VERSION_ID": "12.04"}


def test_ensure_os_compatible(fake_executor, fake_process, mock_get_os_release):
    """Do nothing if the OS is compatible."""
    base_config = ubuntu.BuilddBase(alias=ubuntu.BuilddBaseAlias.JAMMY)

    base_config._ensure_os_compatible(executor=fake_executor)

    mock_get_os_release.assert_called_once()


def test_ensure_os_compatible_devel_mismatch(
    fake_executor, fake_process, logs, mock_get_os_release
):
    """Ignore OS version id mismatch when using a devel base."""
    base_config = ubuntu.BuilddBase(alias=ubuntu.BuilddBaseAlias.DEVEL)

    base_config._ensure_os_compatible(executor=fake_executor)

    mock_get_os_release.assert_called_once()

    assert (
        "Ignoring OS version mismatch for '22.04' because base is 'devel'."
        in logs.debug
    )


def test_ensure_os_compatible_name_failure(
    fake_executor, fake_process, mock_get_os_release
):
    """Raise an error if the OS name does not match."""
    mock_get_os_release.return_value = {"NAME": "Fedora", "VERSION_ID": "32"}
    base_config = ubuntu.BuilddBase(alias=ubuntu.BuilddBaseAlias.JAMMY)

    with pytest.raises(BaseCompatibilityError) as exc_info:
        base_config._ensure_os_compatible(executor=fake_executor)

    assert exc_info.value == BaseCompatibilityError(
        "Expected OS 'Ubuntu', found 'Fedora'"
    )

    mock_get_os_release.assert_called_once()


def test_ensure_os_compatible_version_failure(
    fake_executor, fake_process, mock_get_os_release
):
    """Raise an error if the OS version id does not match."""
    mock_get_os_release.return_value = {"NAME": "Ubuntu", "VERSION_ID": "12.04"}
    base_config = ubuntu.BuilddBase(alias=ubuntu.BuilddBaseAlias.JAMMY)

    with pytest.raises(BaseCompatibilityError) as exc_info:
        base_config._ensure_os_compatible(executor=fake_executor)

    assert exc_info.value == BaseCompatibilityError(
        "Expected OS version '22.04', found '12.04'"
    )

    mock_get_os_release.assert_called_once()


def test_setup_hostname_failure(fake_process, fake_executor):
    base_config = ubuntu.BuilddBase(alias=ubuntu.BuilddBaseAlias.JAMMY)
    fake_process.register_subprocess(
        [*DEFAULT_FAKE_CMD, "hostname", "-F", "/etc/hostname"],
        returncode=-1,
    )

    with pytest.raises(BaseConfigurationError, match="Failed to set hostname."):
        base_config.setup_hostname(executor=fake_executor)


def test_setup_networkd_enable_failure(fake_process, fake_executor):
    base_config = ubuntu.BuilddBase(alias=ubuntu.BuilddBaseAlias.JAMMY)
    fake_process.register_subprocess(
        [*DEFAULT_FAKE_CMD, "systemctl", "enable", "systemd-networkd"],
        returncode=-1,
    )

    with pytest.raises(
        BaseConfigurationError, match="Failed to setup systemd-networkd."
    ):
        base_config._setup_networkd(executor=fake_executor)


def test_setup_networkd_restart_failure(fake_process, fake_executor):
    base_config = ubuntu.BuilddBase(alias=ubuntu.BuilddBaseAlias.JAMMY)
    fake_process.register_subprocess(
        [*DEFAULT_FAKE_CMD, "systemctl", "enable", "systemd-networkd"],
    )
    fake_process.register_subprocess(
        [*DEFAULT_FAKE_CMD, "systemctl", "restart", "systemd-networkd"],
        returncode=-1,
    )

    with pytest.raises(
        BaseConfigurationError, match="Failed to setup systemd-networkd."
    ):
        base_config._setup_networkd(executor=fake_executor)


def test_setup_resolved_enable_failure(fake_process, fake_executor):
    base_config = ubuntu.BuilddBase(alias=ubuntu.BuilddBaseAlias.JAMMY)
    fake_process.register_subprocess(
        [
            *DEFAULT_FAKE_CMD,
            "ln",
            "-sf",
            "/run/systemd/resolve/resolv.conf",
            "/etc/resolv.conf",
        ],
    )
    fake_process.register_subprocess(
        [*DEFAULT_FAKE_CMD, "systemctl", "enable", "systemd-resolved"],
        returncode=-1,
    )

    with pytest.raises(
        BaseConfigurationError, match="Failed to setup systemd-resolved."
    ):
        base_config._setup_resolved(executor=fake_executor)


def test_setup_resolved_restart_failure(fake_process, fake_executor):
    base_config = ubuntu.BuilddBase(alias=ubuntu.BuilddBaseAlias.JAMMY)
    fake_process.register_subprocess(
        [
            *DEFAULT_FAKE_CMD,
            "ln",
            "-sf",
            "/run/systemd/resolve/resolv.conf",
            "/etc/resolv.conf",
        ],
    )
    fake_process.register_subprocess(
        [*DEFAULT_FAKE_CMD, "systemctl", "enable", "systemd-resolved"],
    )
    fake_process.register_subprocess(
        [*DEFAULT_FAKE_CMD, "systemctl", "restart", "systemd-resolved"],
        returncode=-1,
    )

    with pytest.raises(
        BaseConfigurationError, match="Failed to setup systemd-resolved."
    ):
        base_config._setup_resolved(executor=fake_executor)


def test_setup_snapd_proxy(fake_executor, fake_process):
    """Verify snapd proxy is set or unset."""
    environment: dict[str, str | None] = {
        "http_proxy": "http://foo.bar:8080",
        "https_proxy": "http://foo.bar:8081",
    }
    base_config = ubuntu.BuilddBase(
        alias=ubuntu.BuilddBaseAlias.JAMMY,
        environment=environment,
    )
    fake_process.keep_last_process(keep=True)
    fake_process.register([fake_process.any()])

    base_config._setup_snapd_proxy(executor=fake_executor)
    assert [
        *DEFAULT_FAKE_CMD,
        "snap",
        "set",
        "system",
        "proxy.http=http://foo.bar:8080",
    ] in fake_process.calls
    assert [
        *DEFAULT_FAKE_CMD,
        "snap",
        "set",
        "system",
        "proxy.https=http://foo.bar:8081",
    ] in fake_process.calls


@pytest.mark.parametrize("fail_index", list(range(0, 1)))
def test_setup_snapd_proxy_failures(fake_process, fake_executor, fail_index):
    base_config = ubuntu.BuilddBase(alias=ubuntu.BuilddBaseAlias.JAMMY)

    return_codes = [0, 0]
    return_codes[fail_index] = 1

    fake_process.register_subprocess(
        [*DEFAULT_FAKE_CMD, "snap", "unset", "system", "proxy.http"],
        returncode=return_codes[0],
    )
    fake_process.register_subprocess(
        [*DEFAULT_FAKE_CMD, "snap", "unset", "system", "proxy.https"],
        returncode=return_codes[1],
    )

    with pytest.raises(BaseConfigurationError, match="Failed to set the snapd proxy."):
        base_config._setup_snapd_proxy(executor=fake_executor)


@pytest.mark.usefixtures("stub_verify_network")
@pytest.mark.parametrize("fail_index", list(range(0, 2)))
def test_pre_setup_snapd_failures(fake_process, fake_executor, fail_index):
    base_config = ubuntu.BuilddBase(alias=ubuntu.BuilddBaseAlias.JAMMY)

    return_codes = [0, 0]
    return_codes[fail_index] = 1

    fake_process.register_subprocess(
        [*DEFAULT_FAKE_CMD, "systemctl", "is-active", "systemd-udevd"],
        stdout="inactive",
    )
    fake_process.register_subprocess(
        [*DEFAULT_FAKE_CMD, "systemctl", "enable", "systemd-udevd"],
        returncode=return_codes[0],
    )
    fake_process.register_subprocess(
        [*DEFAULT_FAKE_CMD, "systemctl", "start", "systemd-udevd"],
        returncode=return_codes[1],
    )

    with pytest.raises(
        BaseConfigurationError, match="Failed to enable systemd-udevd service."
    ):
        base_config._pre_setup_snapd(executor=fake_executor)


@pytest.mark.usefixtures("stub_verify_network")
def test_setup_snapd_failures(fake_process, fake_executor):
    base_config = ubuntu.BuilddBase(alias=ubuntu.BuilddBaseAlias.JAMMY)
    fake_process.register_subprocess(
        [*DEFAULT_FAKE_CMD, "apt-get", "install", "-y", "snapd"],
        returncode=1,
    )

    with pytest.raises(BaseConfigurationError, match="Failed to setup snapd."):
        base_config._setup_snapd(executor=fake_executor)


@pytest.mark.parametrize("fail_index", list(range(0, 8)))
def test_post_setup_snapd_failures(fake_process, fake_executor, fail_index):
    base_config = ubuntu.BuilddBase(alias=ubuntu.BuilddBaseAlias.JAMMY)
    return_codes = [0] * 8
    return_codes[fail_index] = 1

    fake_process.register_subprocess(
        [*DEFAULT_FAKE_CMD, "ln", "-sf", "/var/lib/snapd/snap", "/snap"],
        returncode=return_codes[0],
    )
    fake_process.register_subprocess(
        [*DEFAULT_FAKE_CMD, "systemctl", "enable", "--now", "snapd.socket"],
        returncode=return_codes[1],
    )
    fake_process.register_subprocess(
        [*DEFAULT_FAKE_CMD, "systemctl", "restart", "snapd.service"],
        returncode=return_codes[2],
    )
    fake_process.register_subprocess(
        [*DEFAULT_FAKE_CMD, "snap", "wait", "system", "seed.loaded"],
        returncode=return_codes[3],
    )
    fake_process.register_subprocess(
        [*DEFAULT_FAKE_CMD, "snap", "refresh", "--hold"],
        returncode=return_codes[4],
    )
    fake_process.register_subprocess(
        [*DEFAULT_FAKE_CMD, "snap", "watch", "--last=auto-refresh?"],
        returncode=return_codes[5],
    )
    fake_process.register_subprocess(
        [*DEFAULT_FAKE_CMD, "snap", "unset", "system", "proxy.http"],
        returncode=return_codes[6],
    )
    fake_process.register_subprocess(
        [*DEFAULT_FAKE_CMD, "snap", "unset", "system", "proxy.https"],
        returncode=return_codes[7],
    )

    with pytest.raises(BaseConfigurationError):
        base_config._post_setup_snapd(executor=fake_executor)


@pytest.mark.parametrize("fail_index", list(range(0, 2)))
def test_warmup_snapd_failures(fake_process, fake_executor, fail_index):
    base_config = ubuntu.BuilddBase(alias=ubuntu.BuilddBaseAlias.JAMMY)
    return_codes = [0] * 2
    return_codes[fail_index] = 1

    fake_process.register_subprocess(
        [*DEFAULT_FAKE_CMD, "snap", "unset", "system", "proxy.http"],
        returncode=return_codes[0],
    )
    fake_process.register_subprocess(
        [*DEFAULT_FAKE_CMD, "snap", "unset", "system", "proxy.https"],
        returncode=return_codes[1],
    )

    with pytest.raises(BaseConfigurationError, match="Failed to set the snapd proxy."):
        base_config._warmup_snapd(executor=fake_executor)


@pytest.mark.parametrize("alias", list(ubuntu.BuilddBaseAlias))
@pytest.mark.parametrize("system_running_ready_stdout", ["degraded", "running"])
def test_wait_for_system_ready(
    fake_executor, fake_process, alias, system_running_ready_stdout
):
    base_config = ubuntu.BuilddBase(alias=alias)
    base_config._retry_wait = 0.01
    fake_process.register_subprocess(
        [*DEFAULT_FAKE_CMD, "systemctl", "is-system-running"],
        stdout="not-ready",
        returncode=-1,
    )
    fake_process.register_subprocess(
        [*DEFAULT_FAKE_CMD, "systemctl", "is-system-running"], stdout="still-not-ready"
    )
    fake_process.register_subprocess(
        [*DEFAULT_FAKE_CMD, "systemctl", "is-system-running"],
        stdout=system_running_ready_stdout,
    )
    fake_process.register_subprocess(
        [*DEFAULT_FAKE_CMD, "getent", "hosts", "snapcraft.io"],
        returncode=-1,
    )
    fake_process.register_subprocess(
        [*DEFAULT_FAKE_CMD, "getent", "hosts", "snapcraft.io"],
        returncode=0,
    )

    base_config.wait_until_ready(executor=fake_executor)

    assert fake_executor.records_of_push_file_io == []
    assert fake_executor.records_of_pull_file == []
    assert fake_executor.records_of_push_file == []
    assert list(fake_process.calls) == [
        [
            *DEFAULT_FAKE_CMD,
            "systemctl",
            "is-system-running",
        ],
        [
            *DEFAULT_FAKE_CMD,
            "systemctl",
            "is-system-running",
        ],
        [
            *DEFAULT_FAKE_CMD,
            "systemctl",
            "is-system-running",
        ],
        [
            *DEFAULT_FAKE_CMD,
            "getent",
            "hosts",
            "snapcraft.io",
        ],
        [
            *DEFAULT_FAKE_CMD,
            "getent",
            "hosts",
            "snapcraft.io",
        ],
    ]


@pytest.mark.parametrize("alias", list(ubuntu.BuilddBaseAlias))
def test_wait_for_system_ready_timeout(fake_executor, fake_process, alias):
    base_config = ubuntu.BuilddBase(
        alias=alias,
    )
    base_config._timeout_simple = 0.01
    base_config._retry_wait = 0.01
    fake_process.register_subprocess(
        [*DEFAULT_FAKE_CMD, "systemctl", "is-system-running"],
        stdout="not-ready",
        returncode=-1,
    )

    with pytest.raises(
        BaseConfigurationError, match="Timed out waiting for environment to be ready."
    ):
        base_config.wait_until_ready(executor=fake_executor)


@pytest.mark.parametrize("alias", list(ubuntu.BuilddBaseAlias))
def test_wait_for_system_ready_timeout_in_network(
    fake_executor, fake_process, alias, monkeypatch
):
    base_config = ubuntu.BuilddBase(alias=alias)
    base_config._timeout_simple = 0.01
    base_config._retry_wait = 0.01
    monkeypatch.setattr(
        base_config, "_setup_wait_for_system_ready", lambda **kwargs: None
    )

    fake_process.register_subprocess(
        [*DEFAULT_FAKE_CMD, "getent", "hosts", "snapcraft.io"],
        returncode=-1,
    )

    with pytest.raises(
        BaseConfigurationError, match="Timed out waiting for networking to be ready."
    ):
        base_config.wait_until_ready(executor=fake_executor)


def test_update_compatibility_tag(fake_executor, mock_load):
    """`update_compatibility_tag()` should update the instance config."""
    base_config = ubuntu.BuilddBase(
        alias=ubuntu.BuilddBaseAlias.JAMMY, compatibility_tag="test-tag"
    )
    base_config._update_compatibility_tag(executor=fake_executor)

    assert fake_executor.records_of_push_file_io == [
        {
            "content": b"compatibility_tag: test-tag\n",
            "destination": "/etc/craft-instance.conf",
            "file_mode": "0644",
            "group": "root",
            "user": "root",
        }
    ]


@pytest.mark.parametrize("status", [True, False])
def test_update_setup_status(fake_executor, mock_load, status):
    """`update_setup_status()` should update the instance config."""
    base_config = ubuntu.BuilddBase(alias=ubuntu.BuilddBaseAlias.JAMMY)

    base_config._update_setup_status(
        executor=fake_executor,
        status=status,
    )

    assert fake_executor.records_of_push_file_io == [
        {
            "content": (
                "compatibility_tag: buildd-base-v7\n"
                f"setup: {str(status).lower()}\n".encode()
            ),
            "destination": "/etc/craft-instance.conf",
            "file_mode": "0644",
            "group": "root",
            "user": "root",
        }
    ]


def test_ensure_config_compatible_validation_error(
    fake_executor, mock_load, fake_validation_error
):
    mock_load.side_effect = fake_validation_error

    base_config = ubuntu.BuilddBase(alias=ubuntu.BuilddBaseAlias.JAMMY)

    with pytest.raises(
        BaseConfigurationError, match="Failed to parse instance configuration file."
    ):
        base_config._ensure_instance_config_compatible(executor=fake_executor)


def test_ensure_config_compatible_empty_config_returns_none(fake_executor, mock_load):
    mock_load.return_value = None

    base_config = ubuntu.BuilddBase(alias=ubuntu.BuilddBaseAlias.JAMMY)

    assert (
        base_config._ensure_instance_config_compatible(executor=fake_executor) is None
    )


def test_ensure_setup_completed(fake_executor, logs, mock_load):
    """Verify the setup was completed by checking the instance config file."""
    mock_load.return_value = InstanceConfiguration(setup=True)

    base_config = ubuntu.BuilddBase(alias=ubuntu.BuilddBaseAlias.JAMMY)

    assert base_config._ensure_setup_completed(executor=fake_executor) is None

    assert "Instance has already been setup." in logs.debug


def test_ensure_setup_completed_validation_error(
    fake_executor, mock_load, fake_validation_error
):
    """Raise an error when the instance config cannot be loaded."""
    mock_load.side_effect = fake_validation_error

    base_config = ubuntu.BuilddBase(alias=ubuntu.BuilddBaseAlias.JAMMY)

    with pytest.raises(
        BaseCompatibilityError, match="failed to parse instance configuration file"
    ):
        base_config._ensure_setup_completed(executor=fake_executor)


def test_ensure_setup_completed_file_not_found_error(fake_executor, mock_load):
    """Raise an error when the instance config cannot be loaded."""
    mock_load.side_effect = FileNotFoundError

    base_config = ubuntu.BuilddBase(alias=ubuntu.BuilddBaseAlias.JAMMY)

    with pytest.raises(BaseCompatibilityError) as raised:
        base_config._ensure_setup_completed(executor=fake_executor)

    assert raised.value == BaseCompatibilityError("failed to find instance config file")


def test_ensure_setup_completed_empty_config(fake_executor, mock_load):
    """Raise an error if the instance config is empty."""
    mock_load.return_value = None

    base_config = ubuntu.BuilddBase(alias=ubuntu.BuilddBaseAlias.JAMMY)

    with pytest.raises(BaseCompatibilityError) as raised:
        base_config._ensure_setup_completed(executor=fake_executor)

    assert raised.value == BaseCompatibilityError("instance config is empty")


@pytest.mark.parametrize("status", [None, False])
def test_ensure_setup_completed_not_setup(status, fake_executor, mock_load):
    """Raise an error if the setup was not completed (setup field is None or False)."""
    mock_load.return_value = InstanceConfiguration(setup=status)

    base_config = ubuntu.BuilddBase(alias=ubuntu.BuilddBaseAlias.JAMMY)

    with pytest.raises(BaseCompatibilityError) as raised:
        base_config._ensure_setup_completed(executor=fake_executor)

    assert raised.value == BaseCompatibilityError("instance is marked as not setup")


@pytest.mark.parametrize(
    "environment",
    [
        None,
        {
            "https_proxy": "http://foo.bar:8081",
            "http_proxy": "http://foo.bar:8080",
        },
    ],
)
@pytest.mark.parametrize("cache_path", [None, pathlib.Path("/tmp")])
def test_warmup_overall(
    environment, fake_process, fake_executor, mock_load, mocker, cache_path
):
    mock_load.return_value = InstanceConfiguration(
        compatibility_tag="buildd-base-v7", setup=True
    )

    alias = ubuntu.BuilddBaseAlias.JAMMY

    if environment is None:
        environment = ubuntu.BuilddBase.default_command_environment()

    base_config = ubuntu.BuilddBase(alias=alias, environment=environment)

    fake_process.register_subprocess(
        [*DEFAULT_FAKE_CMD, "cat", "/etc/os-release"],
        stdout=dedent(
            f"""\
            NAME="Ubuntu"
            ID=ubuntu
            ID_LIKE=debian
            VERSION_ID="{alias.value}"
            """
        ),
    )
    fake_process.register_subprocess(
        [*DEFAULT_FAKE_CMD, "systemctl", "is-system-running"], stdout="degraded"
    )
    fake_process.register_subprocess([*DEFAULT_FAKE_CMD, "chmod", "go+x", "/root"])
    fake_process.register_subprocess(
        [*DEFAULT_FAKE_CMD, "getent", "hosts", "snapcraft.io"]
    )
    fake_process.register_subprocess(
        [*DEFAULT_FAKE_CMD, "snap", "set", "system", "proxy.http=http://foo.bar:8080"]
    )
    fake_process.register_subprocess(
        [*DEFAULT_FAKE_CMD, "snap", "unset", "system", "proxy.http"]
    )
    fake_process.register_subprocess(
        [*DEFAULT_FAKE_CMD, "snap", "set", "system", "proxy.https=http://foo.bar:8081"]
    )
    fake_process.register_subprocess(
        [*DEFAULT_FAKE_CMD, "snap", "unset", "system", "proxy.https"]
    )

    base_config.warmup(executor=fake_executor)

    assert fake_executor.records_of_push_file_io == []
    assert fake_executor.records_of_pull_file == []
    assert fake_executor.records_of_push_file == []


def test_warmup_bad_os(fake_process, fake_executor, mock_load):
    mock_load.return_value = InstanceConfiguration(
        compatibility_tag="buildd-base-v7", setup=True
    )
    base_config = ubuntu.BuilddBase(
        alias=ubuntu.BuilddBaseAlias.JAMMY,
        environment=ubuntu.BuilddBase.default_command_environment(),
    )

    fake_process.register_subprocess(
        [*DEFAULT_FAKE_CMD, "cat", "/etc/os-release"],
        stdout=dedent(
            """\
            NAME="D.O.S."
            ID=dos
            ID_LIKE=dos
            VERSION_ID="DOS 5.1"
            """
        ),
    )

    with pytest.raises(BaseCompatibilityError):
        base_config.warmup(executor=fake_executor)


def test_warmup_bad_instance_config(fake_process, fake_executor, mock_load):
    mock_load.return_value = InstanceConfiguration(
        compatibility_tag="buildd-base-v7", setup=True
    )
    alias = ubuntu.BuilddBaseAlias.JAMMY
    base_config = ubuntu.BuilddBase(
        alias=alias,
        environment=ubuntu.BuilddBase.default_command_environment(),
    )
    base_config.compatibility_tag = "different-tag"

    fake_process.register_subprocess(
        [*DEFAULT_FAKE_CMD, "cat", "/etc/os-release"],
        stdout=dedent(
            f"""\
            NAME="Ubuntu"
            ID=ubuntu
            ID_LIKE=debian
            VERSION_ID="{alias.value}"
            """
        ),
    )

    with pytest.raises(BaseCompatibilityError):
        base_config.warmup(executor=fake_executor)


@pytest.mark.parametrize("setup", [False, None])
def test_warmup_not_setup(setup, fake_process, fake_executor, mock_load):
    """Raise a BaseConfigurationError if the instance is not setup."""
    mock_load.return_value = InstanceConfiguration(
        compatibility_tag="buildd-base-v7", setup=setup
    )
    alias = ubuntu.BuilddBaseAlias.JAMMY
    base_config = ubuntu.BuilddBase(
        alias=alias,
        environment=ubuntu.BuilddBase.default_command_environment(),
    )

    fake_process.register_subprocess(
        [*DEFAULT_FAKE_CMD, "cat", "/etc/os-release"],
        stdout=dedent(
            f"""\
            NAME="Ubuntu"
            ID=ubuntu
            ID_LIKE=debian
            VERSION_ID="{alias.value}"
            """
        ),
    )

    with pytest.raises(BaseCompatibilityError) as raised:
        base_config.warmup(executor=fake_executor)

    assert raised.value == BaseCompatibilityError("instance is marked as not setup")


def test_warmup_never_ready(fake_process, fake_executor, mock_load):
    mock_load.return_value = InstanceConfiguration(
        compatibility_tag="buildd-base-v7", setup=True
    )
    alias = ubuntu.BuilddBaseAlias.JAMMY
    base_config = ubuntu.BuilddBase(
        alias=alias,
        environment=ubuntu.BuilddBase.default_command_environment(),
    )

    fake_process.register_subprocess(
        [*DEFAULT_FAKE_CMD, "cat", "/etc/os-release"],
        stdout=dedent(
            f"""\
            NAME="Ubuntu"
            ID=ubuntu
            ID_LIKE=debian
            VERSION_ID="{alias.value}"
            """
        ),
    )
    for _ in range(3):  # it will called twice until timeout, one extra for safety
        fake_process.register_subprocess(
            [*DEFAULT_FAKE_CMD, "systemctl", "is-system-running"],
            stdout="starting",
        )
    base_config._timeout_simple = 0.001

    with pytest.raises(BaseConfigurationError):
        base_config.warmup(executor=fake_executor)


def test_warmup_never_network(fake_process, fake_executor, mock_load):
    mock_load.return_value = InstanceConfiguration(
        compatibility_tag="buildd-base-v7", setup=True
    )
    alias = ubuntu.BuilddBaseAlias.JAMMY
    base_config = ubuntu.BuilddBase(
        alias=alias,
        environment=ubuntu.BuilddBase.default_command_environment(),
    )

    fake_process.register_subprocess(
        [*DEFAULT_FAKE_CMD, "cat", "/etc/os-release"],
        stdout=dedent(
            f"""\
            NAME="Ubuntu"
            ID=ubuntu
            ID_LIKE=debian
            VERSION_ID="{alias.value}"
            """
        ),
    )
    fake_process.register_subprocess(
        [*DEFAULT_FAKE_CMD, "systemctl", "is-system-running"], stdout="degraded"
    )
    for _ in range(3):  # it will called twice until timeout, one extra for safety
        fake_process.register_subprocess(
            [*DEFAULT_FAKE_CMD, "getent", "hosts", "snapcraft.io"], returncode=1
        )

    base_config._timeout_simple = 0.001
    with pytest.raises(BaseConfigurationError):
        base_config.warmup(executor=fake_executor)


@pytest.mark.parametrize(
    "hostname",
    [
        "t",
        "test",
        "test1",
        "test-1",
        "1-test",
        "this-is-40-characters-xxxxxxxxxxxxxxxxxx",
        "this-is-63-characters-xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx",
    ],
)
def test_set_hostname_unchanged(hostname, logs):
    """Verify hostnames that are already compliant are not changed."""
    base_config = ubuntu.BuilddBase(
        alias=ubuntu.BuilddBaseAlias.JAMMY,
        hostname=hostname,
    )

    assert base_config._hostname == hostname
    assert Exact(f"Using hostname '{hostname}'") in logs.debug


@pytest.mark.parametrize(
    ("hostname", "expected_hostname"),
    [
        # trim away invalid beginning characters
        ("-test", "test"),
        # trim away invalid ending characters
        ("test-", "test"),
        ("test--", "test"),
        ("test1-", "test1"),
        # trim away invalid characters
        ("test$", "test"),
        ("test-!@#$%^&*()test", "test-test"),
        ("$1test", "1test"),
        ("test-$", "test"),
        # this name contains invalid characters so it gets converted, even
        # though it is 63 characters
        (
            "this-is-63-characters-with-invalid-characters-$$$xxxxxxxxxxxxxX",
            "this-is-63-characters-with-invalid-characters-xxxxxxxxxxxxxX",
        ),
        # this name is longer than 63 characters, so it gets converted
        (
            "this-is-64-characters-with-valid-characters-xxxxxxxxxxxxxxxxxxXx",
            "this-is-64-characters-with-valid-characters-xxxxxxxxxxxxxxxxxxX",
        ),
        # trim away away invalid characters and truncate to 63 characters
        (
            "-this-is-64-characters-and-has-invalid-characters-$$$xxxxxxxxxx-",
            "this-is-64-characters-and-has-invalid-characters-xxxxxxxxxx",
        ),
        # ensure invalid ending characters are removed after truncating
        (
            "this-is-64-characters-and-has-a-hyphen-at-character-63-xxxxxxx-x",
            "this-is-64-characters-and-has-a-hyphen-at-character-63-xxxxxxx",
        ),
    ],
)
def test_set_hostname(hostname, expected_hostname, logs):
    """Verify hostname is compliant with hostname naming conventions."""
    base_config = ubuntu.BuilddBase(
        alias=ubuntu.BuilddBaseAlias.JAMMY,
        hostname=hostname,
    )

    assert base_config._hostname == expected_hostname
    assert Exact(f"Using hostname '{expected_hostname}'") in logs.debug


@pytest.mark.parametrize(
    "hostname",
    [
        "",
        "-",
        "$$$",
        "-$-$-",
    ],
)
def test_set_hostname_invalid(hostname):
    """Verify invalid hostnames raise an error."""
    with pytest.raises(BaseConfigurationError) as error:
        ubuntu.BuilddBase(alias=ubuntu.BuilddBaseAlias.JAMMY, hostname=hostname)

    assert error.value == BaseConfigurationError(
        brief=f"failed to create base with hostname {hostname!r}.",
        details="hostname must contain at least one alphanumeric character",
    )


def test_execute_run_default(fake_executor):
    """Default _execute_run behaviour."""
    base_config = ubuntu.BuilddBase(alias=ubuntu.BuilddBaseAlias.JAMMY)
    command = ["the", "command"]
    with patch.object(fake_executor, "execute_run") as mock:
        base_config._execute_run(command, executor=fake_executor)

    mock.assert_called_with(
        command, check=True, capture_output=True, text=False, timeout=None
    )


def test_execute_run_options_for_run(fake_executor):
    """Different options to control how run is called."""
    base_config = ubuntu.BuilddBase(alias=ubuntu.BuilddBaseAlias.JAMMY)
    command = ["the", "command"]
    with patch.object(fake_executor, "execute_run") as mock:
        base_config._execute_run(
            command,
            executor=fake_executor,
            check=False,
            capture_output=False,
            timeout=None,
        )

    mock.assert_called_with(
        command, check=False, capture_output=False, text=False, timeout=None
    )


def test_execute_run_command_failed_no_verify_network(fake_process, fake_executor):
    """The command failed but network verification was not asked."""
    command = ["the", "command"]
    base_config = ubuntu.BuilddBase(alias=ubuntu.BuilddBaseAlias.JAMMY)
    fake_process.register_subprocess([*DEFAULT_FAKE_CMD, *command], returncode=1)

    # we know that network is not verified because otherwise we'll get
    # a ProcessNotRegisteredError for the verification process
    with pytest.raises(subprocess.CalledProcessError):
        base_config._execute_run(command, executor=fake_executor)


@pytest.mark.parametrize("proxy_variable_name", ["HTTPS_PROXY", "https_proxy"])
def test_execute_run_command_failed_verify_network_proxy(
    fake_process, fake_executor, monkeypatch, proxy_variable_name
):
    """The command failed, network verification was asked, but there is a proxy."""
    command = ["the", "command"]
    base_config = ubuntu.BuilddBase(alias=ubuntu.BuilddBaseAlias.JAMMY)
    fake_process.register_subprocess([*DEFAULT_FAKE_CMD, *command], returncode=1)

    monkeypatch.setenv(proxy_variable_name, "https://someproxy.net:8080/")

    # we know that network is not verified because otherwise we'll get
    # a ProcessNotRegisteredError for the verification process
    with pytest.raises(subprocess.CalledProcessError):
        base_config._execute_run(command, executor=fake_executor, verify_network=True)


def test_execute_run_verify_network_run_ok(fake_process, fake_executor):
    """Indicated network verification but process completed ok."""
    command = ["the", "command"]
    base_config = ubuntu.BuilddBase(alias=ubuntu.BuilddBaseAlias.JAMMY)
    fake_process.register_subprocess([*DEFAULT_FAKE_CMD, *command], returncode=0)

    # we know that network is not verified because otherwise we'll get
    # a ProcessNotRegisteredError for the verification process
    proc = base_config._execute_run(
        command, executor=fake_executor, verify_network=True
    )
    assert proc.returncode == 0


@pytest.mark.usefixtures("stub_verify_network")
def test_execute_run_verify_network_connectivity_ok(fake_process, fake_executor):
    """Network verified after process failure, connectivity ok."""
    command = ["the", "command"]
    base_config = ubuntu.BuilddBase(alias=ubuntu.BuilddBaseAlias.JAMMY)

    fake_process.register_subprocess([*DEFAULT_FAKE_CMD, *command], returncode=1)

    with pytest.raises(subprocess.CalledProcessError):
        base_config._execute_run(command, executor=fake_executor, verify_network=True)


def test_execute_run_verify_network_connectivity_missing(fake_process, fake_executor):
    """Network verified after process failure, no connectivity."""
    command = ["the", "command"]
    base_config = ubuntu.BuilddBase(alias=ubuntu.BuilddBaseAlias.JAMMY)

    fake_process.register_subprocess(
        [*DEFAULT_FAKE_CMD, "bash", "-c", "exec 3<> /dev/tcp/snapcraft.io/443"],
        returncode=1,
    )
    fake_process.register_subprocess([*DEFAULT_FAKE_CMD, *command], returncode=1)

    with pytest.raises(NetworkError) as exc_info:
        base_config._execute_run(command, executor=fake_executor, verify_network=True)
    assert isinstance(exc_info.value.__cause__, subprocess.CalledProcessError)


def test_execute_run_bad_check_verifynetwork_combination(fake_executor):
    """Cannot ask for network verification and avoid checking."""
    base_config = ubuntu.BuilddBase(alias=ubuntu.BuilddBaseAlias.JAMMY)
    with pytest.raises(RuntimeError):
        base_config._execute_run(
            ["cmd"], executor=fake_executor, check=False, verify_network=True
        )


@pytest.mark.usefixtures("stub_verify_network")
def test_network_connectivity_yes(fake_executor, fake_process):
    """Connectivity is ok."""
    base_config = ubuntu.BuilddBase(alias=ubuntu.BuilddBaseAlias.JAMMY)

    assert base_config._network_connected(executor=fake_executor) is True


def test_network_connectivity_no(fake_executor, fake_process):
    """Connectivity missing."""
    base_config = ubuntu.BuilddBase(alias=ubuntu.BuilddBaseAlias.JAMMY)
    fake_process.register_subprocess(
        [*DEFAULT_FAKE_CMD, "bash", "-c", "exec 3<> /dev/tcp/snapcraft.io/443"],
        returncode=1,
    )
    assert base_config._network_connected(executor=fake_executor) is False


def test_network_connectivity_timeouts(fake_executor, fake_process):
    """Check that timeout is used.

    This test does not register the fake subprocess with a long wait because to make it
    resilient to CIs it would need a too long waiting.
    """
    base_config = ubuntu.BuilddBase(alias=ubuntu.BuilddBaseAlias.JAMMY)
    cmd = ["bash", "-c", "exec 3<> /dev/tcp/snapcraft.io/443"]
    timeout_expired = subprocess.TimeoutExpired(cmd, timeout=5)
    with patch.object(
        fake_executor, "execute_run", side_effect=timeout_expired
    ) as mock:
        assert base_config._network_connected(executor=fake_executor) is False
    mock.assert_called_with(cmd, check=False, capture_output=True, timeout=10)


def test_disable_and_wait_for_snap_refresh_hold_error(fake_process, fake_executor):
    """Raise BaseConfigurationError when the command to hold snap refreshes fails."""
    base_config = ubuntu.BuilddBase(alias=ubuntu.BuilddBaseAlias.JAMMY)
    fake_process.register_subprocess(
        [*DEFAULT_FAKE_CMD, "snap", "refresh", "--hold"],
        returncode=-1,
    )

    with pytest.raises(BaseConfigurationError, match="Failed to hold snap refreshes."):
        base_config._disable_and_wait_for_snap_refresh(executor=fake_executor)


def test_disable_and_wait_for_snap_refresh_wait_error(fake_process, fake_executor):
    """Raise BaseConfigurationError when the `snap watch` command fails."""
    base_config = ubuntu.BuilddBase(alias=ubuntu.BuilddBaseAlias.JAMMY)
    fake_process.register_subprocess([*DEFAULT_FAKE_CMD, "snap", "refresh", "--hold"])
    fake_process.register_subprocess(
        [*DEFAULT_FAKE_CMD, "snap", "watch", "--last=auto-refresh?"],
        returncode=-1,
    )

    with pytest.raises(
        BaseConfigurationError, match="Failed to wait for snap refreshes to complete."
    ):
        base_config._disable_and_wait_for_snap_refresh(executor=fake_executor)


def test_devel_never_eol(fake_process, fake_executor, mock_requests_head, logs):
    """Verify that the devel base skips EOL checks."""
    base_config = ubuntu.BuilddBase(alias=ubuntu.BuilddBaseAlias.DEVEL)
    fake_process.register_subprocess(
        [*DEFAULT_FAKE_CMD, "cat", "/etc/os-release"],
        stdout="UBUNTU_CODENAME=grumpy",
    )

    base_config._update_eol_sources(fake_executor)

    assert re.escape("Skipping EOL check for base 'devel'.") in logs.debug


@freeze_time("2027-01-01")
@pytest.mark.parametrize("mock_requests_head", [200], indirect=True)
def test_base_past_eol(fake_process, fake_executor, mock_requests_head, logs):
    """Update the sources if the base is past its EOL date."""
    base_config = ubuntu.BuilddBase(alias=ubuntu.BuilddBaseAlias.PLUCKY)
    fake_process.register_subprocess(
        [*DEFAULT_FAKE_CMD, "cat", "/etc/os-release"],
        stdout="UBUNTU_CODENAME=plucky",
    )
    fake_process.register_subprocess(
        [*DEFAULT_FAKE_CMD, "bash", "/tmp/craft-sources.sh"]
    )

    base_config._update_eol_sources(fake_executor)

    assert re.escape("Getting EOL data for 25.04 (plucky).") in logs.debug
    assert re.escape("25.04 is EOL.") in logs.debug
    assert (
        re.escape("Checking for 25.04 (plucky) on https://old-releases.ubuntu.com.")
        in logs.debug
    )
    assert (
        re.escape("25.04 is available on https://old-releases.ubuntu.com.")
        in logs.debug
    )
    assert re.escape("Updating EOL sources.") in logs.debug
    with importlib.resources.path(
        "craft_providers.util", "sources.sh"
    ) as sources_script:
        assert fake_executor.records_of_push_file == [
            {"source": sources_script, "destination": Path("/tmp/craft-sources.sh")}
        ]


@freeze_time("2025-01-01")
@pytest.mark.parametrize("mock_requests_head", [200], indirect=True)
def test_base_not_past_eol(fake_process, fake_executor, mock_requests_head, logs):
    """Don't update EOL sources if the base isn't past its EOL date."""
    base_config = ubuntu.BuilddBase(alias=ubuntu.BuilddBaseAlias.PLUCKY)

    fake_process.register_subprocess(
        [*DEFAULT_FAKE_CMD, "cat", "/etc/os-release"],
        stdout="UBUNTU_CODENAME=plucky",
    )

    base_config._update_eol_sources(fake_executor)

    assert re.escape("Not updating EOL sources because 25.04 isn't EOL.") in logs.debug


@freeze_time("2027-01-01")
def test_base_not_on_old_releases(
    fake_process, fake_executor, mock_requests_head, logs
):
    """Don't update EOL sources if the base isn't on old-releases.ubuntu.com."""
    base_config = ubuntu.BuilddBase(alias=ubuntu.BuilddBaseAlias.PLUCKY)
    fake_process.register_subprocess(
        [*DEFAULT_FAKE_CMD, "cat", "/etc/os-release"],
        stdout="UBUNTU_CODENAME=plucky",
    )

    base_config._update_eol_sources(fake_executor)

    assert (
        re.escape(
            "Not updating EOL sources because 25.04 isn't on https://old-releases.ubuntu.com."
        )
        in logs.debug
    )


@freeze_time("2027-01-01")
def test_get_old_releases_retry(fake_process, fake_executor, mock_requests_head, logs):
    """Retry requests to https://old-releases.ubuntu.com."""
    base_config = ubuntu.BuilddBase(alias=ubuntu.BuilddBaseAlias.PLUCKY)
    mock_requests_head.side_effect = [
        requests.ConnectionError(),
        requests.ConnectionError,
        mock_requests_head.return_value,
    ]
    fake_process.register_subprocess(
        [*DEFAULT_FAKE_CMD, "cat", "/etc/os-release"],
        stdout="UBUNTU_CODENAME=plucky",
    )

    base_config._update_eol_sources(fake_executor)

    assert (
        re.escape(
            "Not updating EOL sources because 25.04 isn't on https://old-releases.ubuntu.com."
        )
        in logs.debug
    )


@freeze_time("2027-01-01")
def test_get_old_releases_error(fake_process, fake_executor, mock_requests_head, logs):
    """Error if https://old-releases.ubuntu.com can't be reached."""
    base_config = ubuntu.BuilddBase(alias=ubuntu.BuilddBaseAlias.PLUCKY)
    mock_requests_head.side_effect = requests.ConnectionError()
    fake_process.register_subprocess(
        [*DEFAULT_FAKE_CMD, "cat", "/etc/os-release"],
        stdout="UBUNTU_CODENAME=plucky",
    )
    # shorten the retries because mock_instant_sleep doesn't work well with freeze_time
    base_config._retry_wait = 0.01
    base_config._timeout_simple = 0.001

    expected_error = re.escape(
        "Failed to get https://old-releases.ubuntu.com/ubuntu/dists/plucky/."
    )
    with pytest.raises(BaseConfigurationError, match=expected_error):
        base_config._update_eol_sources(fake_executor)


def test_get_codename_error(fake_process, fake_executor, mock_requests_head, logs):
    """Error if the codename isn't in /etc/os-release."""
    base_config = ubuntu.BuilddBase(alias=ubuntu.BuilddBaseAlias.PLUCKY)

    fake_process.register_subprocess(
        [*DEFAULT_FAKE_CMD, "cat", "/etc/os-release"],
        stdout="OTHER=value",
    )

    with pytest.raises(BaseConfigurationError) as raised:
        base_config._update_eol_sources(fake_executor)

    assert raised.value == BaseConfigurationError(
        brief="Couldn't find Ubuntu codename in OS release data.",
        details="OS release data: {'OTHER': 'value'}.",
    )


@freeze_time("2027-01-01")
@pytest.mark.parametrize("mock_requests_head", [200], indirect=True)
def test_update_eol_sources_error(
    fake_process, fake_executor, mock_requests_head, logs
):
    """Error if the bash script fails to run in the instance."""

    def _raise_error(process) -> None:
        raise subprocess.CalledProcessError(100, ["error"])

    base_config = ubuntu.BuilddBase(alias=ubuntu.BuilddBaseAlias.PLUCKY)

    fake_process.register_subprocess(
        [*DEFAULT_FAKE_CMD, "cat", "/etc/os-release"],
        stdout="UBUNTU_CODENAME=plucky",
    )
    fake_process.register_subprocess(
        [*DEFAULT_FAKE_CMD, "bash", "/tmp/craft-sources.sh"],
        callback=_raise_error,
    )

    with pytest.raises(BaseConfigurationError) as raised:
        base_config._update_eol_sources(fake_executor)

    assert raised.value == BaseConfigurationError(
        brief="Failed to update EOL sources.",
        details="* Command that failed: 'error'\n* Command exit code: 100",
    )



# Copyright 2025 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License version 3, as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranties of MERCHANTABILITY,
# SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.

import stat
from pathlib import Path

import pytest
from craft_providers import errors, lxd
from craft_providers.bases import ubuntu
from craft_providers.provider import Provider


@pytest.mark.slow
def test_ubuntu_eol_sources(installed_lxd, instance_name, session_lxd_project, mocker):
    """Update EOL sources when setting up an Ubuntu instance."""
    # There are no working EOL buildd bases, so mock the base and use the error to validate the sources were updated.
    mocker.patch.object(ubuntu.BuilddBase, "_get_codename", return_value="mantic")
    base_configuration = ubuntu.BuilddBase(alias=ubuntu.BuilddBaseAlias.NOBLE)
    instance = lxd.LXDInstance(name="test-instance", project=session_lxd_project)

    try:
        with pytest.raises(errors.BaseConfigurationError) as raised:
            lxd.launch(
                name="test-instance",
                project=session_lxd_project,
                base_configuration=base_configuration,
                image_name="24.04",
                image_remote="ubuntu",
                ephemeral=True,
            )

        assert (
            "The repository 'http://old-releases.ubuntu.com/ubuntu noble Release' does not have a Release file."
            in str(raised.value.details)
        )

    finally:
        instance.delete(force=True)


@pytest.mark.slow
def test_root_perms(session_provider: Provider, tmp_path: Path) -> None:
    with session_provider.launched_environment(
        project_name="test",
        instance_name="test-instance",
        project_path=tmp_path,
        base_configuration=ubuntu.BuilddBase(alias=ubuntu.BuilddBaseAlias.NOBLE),
    ) as prov:
        proc = prov.execute_run(
            command=["stat", "-c", "%a", "/root"],
            capture_output=True,
        )

    expected_perms = stat.S_IXGRP | stat.S_IXOTH
    assert int(proc.stdout, base=8) & expected_perms == expected_perms



#
# Copyright 2021-2023 Canonical Ltd.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#

"""Parser for /etc/os-release."""

from __future__ import annotations

import shlex
from pathlib import Path

OS_RELEASE_FILE = Path("/etc/os-release")


def parse_os_release(content: str | None = None) -> dict[str, str]:
    """Parser for /etc/os-release.

    Format documentation at:

    https://www.freedesktop.org/software/systemd/man/os-release.html

    :param content: String contents of os-release file.  If None, will read contents of
    file from host.

    :returns: Dictionary of key-mappings found in os-release. Values are
    stripped of encapsulating quotes.
    """
    if content is None:
        content = OS_RELEASE_FILE.read_text()

    mappings: dict[str, str] = {}

    for line in shlex.split(content):
        key, eq, value = line.partition("=")
        if eq != "=":  # Not a variable getting set; ignore.
            continue
        mappings[key] = value

    return mappings



#
# Copyright 2025 Canonical Ltd.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
import contextlib
import textwrap
from unittest.mock import MagicMock, patch

import pytest
from craft_providers.bases import centos, ensure_guest_compatible, ubuntu
from craft_providers.errors import ProviderError
from craft_providers.util import os_release

from tests.unit.conftest import DEFAULT_FAKE_CMD


def test_ensure_guest_compatible_not_ubuntu(fake_executor, fake_process):
    base = centos.CentOSBase(alias=centos.CentOSBaseAlias.SEVEN)
    base.get_os_release = MagicMock(spec=base.get_os_release)
    ensure_guest_compatible(base, fake_executor, "")

    # The first thing that ensure_guest_compatible does is the base check.  The next
    # thing is to call get_os_release on the base.  So if that isn't called then we
    # haven't progressed.
    base.get_os_release.assert_not_called()


def test_ensure_guest_compatible_non_ubuntu_host(
    fake_executor,
    fake_process,
):
    """Check for combinations of host and guest OS unaffected by the lxd issue."""
    guest_base = ubuntu.BuilddBase(alias=ubuntu.BuilddBaseAlias.JAMMY)
    guest_base.get_os_release = MagicMock(spec=guest_base.get_os_release)

    # Mock the host os-release file
    fake_process.register_subprocess(
        [*DEFAULT_FAKE_CMD, "cat", "/etc/os-release"],
        stdout='ID="fedora"',
    )

    # Mock the host os-release file contents
    @contextlib.contextmanager
    def fake_open(*args, **kwargs):
        class Fake:
            def read(self):
                return 'ID="fedora"'

        yield Fake()

    with patch.object(os_release.Path, "open", fake_open):
        ensure_guest_compatible(guest_base, fake_executor, "4.0")

    # The first thing that ensure_guest_compatible does is the base check.  The next
    # thing is to call get_os_release on the base.  So if that isn't called then we
    # haven't progressed.
    guest_base.get_os_release.assert_not_called()


@pytest.mark.parametrize(
    "base_alias",
    [
        ubuntu.BuilddBaseAlias.JAMMY,  # host greater than FOCAL
        ubuntu.BuilddBaseAlias.FOCAL,  # guest less than ORACULAR
    ],
)
def test_ensure_guest_compatible_valid_ubuntu(
    fake_executor,
    fake_process,
    base_alias,
):
    """Check for combinations of host and guest OS unaffected by the lxd issue."""
    guest_base = ubuntu.BuilddBase(alias=base_alias)
    guest_base._retry_wait = 0.01
    guest_base._timeout_simple = 1

    # Set this up so we can be sure the guest get_os_release was called once
    real_get_os_release = guest_base.get_os_release

    counter = 0

    def fake_get_os_release(*args, **kwargs):
        nonlocal counter
        counter += 1
        return real_get_os_release(*args, **kwargs)

    guest_base.get_os_release = fake_get_os_release

    lxd_version = "0.0.0"

    # Mock the host os-release file
    fake_os_release = textwrap.dedent(
        f"""\
        ID="ubuntu"
        VERSION_ID="{base_alias.value}"
        WOOP="dedoo"
        """
    )
    fake_process.register_subprocess(
        [*DEFAULT_FAKE_CMD, "cat", "/etc/os-release"],
        stdout=fake_os_release,
    )

    # Mock the host os-release file contents
    @contextlib.contextmanager
    def fake_open(*args, **kwargs):
        class Fake:
            def read(self):
                return textwrap.dedent(
                    f"""\
                    ID="ubuntu"
                    VERSION_ID="{base_alias.value}"
                    """
                )

        yield Fake()

    # Kernel version doesn't matter for this test, but setting it allows the test to
    # pass on windows
    with (
        patch("platform.release", return_value="4.99"),
        patch.object(os_release.Path, "open", fake_open),
    ):
        ensure_guest_compatible(guest_base, fake_executor, lxd_version)

    assert counter == 1


@pytest.mark.parametrize(
    ("host_base_alias", "guest_base_alias", "lxd_version", "kernel_version"),
    [
        # host FOCAL or older AND guest ORACULAR or newer
        (
            ubuntu.BuilddBaseAlias.BIONIC,
            ubuntu.BuilddBaseAlias.ORACULAR,
            "6.0",
            "5.14",
        ),
        (
            ubuntu.BuilddBaseAlias.FOCAL,
            ubuntu.BuilddBaseAlias.ORACULAR,
            "5.0.4",
            "2.6.32-51555-generic",
        ),
        (
            ubuntu.BuilddBaseAlias.XENIAL,
            ubuntu.BuilddBaseAlias.DEVEL,
            "4.5",
            "16.661",
        ),
    ],
)
def test_ensure_guest_compatible_bad_kernel_versions(
    mocker,
    fake_executor,
    fake_process,
    host_base_alias,
    guest_base_alias,
    lxd_version,
    kernel_version,
):
    """Various kernels must raise when the host and guest OS and LXD versions all match."""
    guest_base = ubuntu.BuilddBase(alias=guest_base_alias)
    guest_base._retry_wait = 0.01
    guest_base._timeout_simple = 1

    # Mock the guest os-release file contents
    fake_process.register_subprocess(
        [*DEFAULT_FAKE_CMD, "cat", "/etc/os-release"],
        stdout=f'VERSION_ID="{guest_base_alias.value}"',
    )

    # Mock the host os-release file contents
    @contextlib.contextmanager
    def fake_open(*args, **kwargs):
        class Fake:
            def read(self):
                return textwrap.dedent(
                    f"""\
                    ID="ubuntu"
                    VERSION_ID="{host_base_alias.value}"
                    """
                )

        yield Fake()

    with (
        patch("platform.release", return_value=kernel_version),
        patch.object(os_release.Path, "open", fake_open),
        pytest.raises(ProviderError) as e,
    ):
        ensure_guest_compatible(guest_base, fake_executor, lxd_version)
    assert (
        "This combination of guest and host OS versions requires a newer kernel and/or lxd."
        in str(e)
    )



#
# Copyright 2021-2023 Canonical Ltd.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#

"""Collection of bases used to configure build environments."""

from enum import Enum
import sys
from typing import Literal, NamedTuple, overload

from craft_providers.errors import BaseCompatibilityError, BaseConfigurationError
from craft_providers.base import Base

from . import almalinux, centos
from . import ubuntu
from . import ubuntu as buildd
from .checks import ensure_guest_compatible
from .ubuntu import BuilddBase, BuilddBaseAlias

sys.modules["craft_providers.bases.buildd"] = buildd

BaseAlias = (
    ubuntu.BuilddBaseAlias | almalinux.AlmaLinuxBaseAlias | centos.CentOSBaseAlias
)

__all__ = [
    "centos",
    "ensure_guest_compatible",
    "ubuntu",
    "BaseAlias",
    "BaseName",
    "BuilddBase",
    "BuilddBaseAlias",
    "BaseCompatibilityError",
    "BaseConfigurationError",
]


class BaseName(NamedTuple):
    """A base image, by distribution and version."""

    name: str
    version: str


BASE_NAME_TO_BASE_ALIAS: dict[BaseName, BaseAlias] = {
    BaseName("ubuntu", "16.04"): ubuntu.BuilddBaseAlias.XENIAL,
    BaseName("ubuntu", "18.04"): ubuntu.BuilddBaseAlias.BIONIC,
    BaseName("ubuntu", "20.04"): ubuntu.BuilddBaseAlias.FOCAL,
    BaseName("ubuntu", "22.04"): ubuntu.BuilddBaseAlias.JAMMY,
    BaseName("ubuntu", "24.04"): ubuntu.BuilddBaseAlias.NOBLE,
    BaseName("ubuntu", "24.10"): ubuntu.BuilddBaseAlias.ORACULAR,
    BaseName("ubuntu", "25.04"): ubuntu.BuilddBaseAlias.PLUCKY,
    BaseName("ubuntu", "25.10"): ubuntu.BuilddBaseAlias.QUESTING,
    BaseName("ubuntu", "26.04"): ubuntu.BuilddBaseAlias.RESOLUTE,
    BaseName("ubuntu", "devel"): ubuntu.BuilddBaseAlias.DEVEL,
    BaseName("centos", "7"): centos.CentOSBaseAlias.SEVEN,
    BaseName("almalinux", "9"): almalinux.AlmaLinuxBaseAlias.NINE,
}


@overload
def get_base_alias(
    base_name: tuple[Literal["ubuntu"], str],
) -> ubuntu.BuilddBaseAlias: ...
@overload
def get_base_alias(
    base_name: tuple[Literal["centos"], str],
) -> centos.CentOSBaseAlias: ...
@overload
def get_base_alias(
    base_name: tuple[Literal["almalinux"], str],
) -> almalinux.AlmaLinuxBaseAlias: ...
@overload
def get_base_alias(base_name: BaseName) -> BaseAlias: ...
def get_base_alias(base_name: tuple[str, str]) -> BaseAlias:
    """Return a Base alias from a base (name, version) tuple."""
    base_name = BaseName(*base_name)
    if base_name.name == "ubuntu" and base_name in BASE_NAME_TO_BASE_ALIAS:
        return BASE_NAME_TO_BASE_ALIAS[base_name]

    # match other distributions sub-versions like 9.1 to 9
    _base_name = BaseName(base_name.name, base_name.version.split(".")[0])
    if _base_name in BASE_NAME_TO_BASE_ALIAS:
        return BASE_NAME_TO_BASE_ALIAS[_base_name]

    raise BaseConfigurationError(f"Base alias not found for {base_name}")


@overload
def get_base_from_alias(alias: ubuntu.BuilddBaseAlias) -> type[ubuntu.BuilddBase]: ...
@overload
def get_base_from_alias(alias: centos.CentOSBaseAlias) -> type[centos.CentOSBase]: ...
@overload
def get_base_from_alias(
    alias: almalinux.AlmaLinuxBaseAlias,
) -> type[almalinux.AlmaLinuxBase]: ...
def get_base_from_alias(alias: BaseAlias) -> type[Base[Enum]]:
    """Return a Base class from a known base alias."""
    match alias:
        case ubuntu.BuilddBaseAlias():
            return ubuntu.BuilddBase
        case centos.CentOSBaseAlias():
            return centos.CentOSBase
        case almalinux.AlmaLinuxBaseAlias():
            return almalinux.AlmaLinuxBase
        case _:
            raise BaseConfigurationError(f"Base not found for alias {alias}")
